/**************************************************************************************************************************
// This script was authored by Lorne Brooks, Email: lbtools2k@gmail.com
//
//
// REVISION HISTORY:
// Version 3.2 13/10/23;
//						- Fixed a crash when attempting to paste a copied posture with an existing empty CopyPosture collection.
//						- Move the Keys section to its own separate rollout.
//						- Added a Time Range section which applies to all functions that operate across the time range.
//						- Key All Parts and Key Selected Parts can now add new keys across the time range by holding either shift or control.
//						- Added a new method to delete keys across the time range for either selected parts, or all parts if shift or control are held.
//						- Added a new method to delete in-between keys across the time range based a step size set by a spinner. Keys that exist between each step will be removed.
//						- Added ability to paste a copied posture across the time range for each frame when holding shift or control.
//						- Added additional options for pasting a posture on the Biped root (COM).
//						- Added a new method to Posture Grabber for straightening fingers with finger selection.
//						- Added left and right hand select all fingers buttons.
//						- Added prop alignment and baking to a target object to the Baking rollout.
//						- Added new methods to Layers rollout for adding, deleting, collapsing and renaming Biped layers.
//						- Added support for tail bones when setting or deleting keys.
//						- Added selection of props if available.
//						- Added checks for invalid biped to all UI functions.
//						- Mirroring limb postures now includes fingers and toes.
//						- Script now handles finding Biped objects with modifiers in the stack.
//						- Various UI and layout adjustments.
// Version 3.1 24/09/22;
//						- Changed PDF guide button to link to an online guide.
//						- Added ability to hold shift to set planted keys for limbs. Store and load spaces now account for planted keys.
//						- IK pivot UI's and layers retargeting UI now follow main UI if moved.
// Version 3.0 23/02/21;
//						- Renamed previous method "Align & Bake To Target" to "Align To Offset".
//						- Added new alignment method "Align To Match" which re-aligns a Biped for all root keys to match a target object's animation (can be another Biped root).
// Version 2.9 21/02/21;
//						- Added a new rollout "Alignment" which contains a method for aligning and baking to a target object for any keys found on the root and limbs.
// Version 2.8 09/11/20;
//						- Added to layers rollout a Snap & Set Key button and a separate dialog for Retargeting.
// Version 2.7 02/11/20;
//						- Added a new rollout for layers navigation. Includes: move up and down layer buttons, a drop-down list of available layers,
//						  and active layer toggles.
// Version 2.6 06/10/20;
//						- Added ability to open a new window to set IK pivots for each hand & foot where planted or sliding keys are assigned.
// Version 2.5 25/09/20;
//						- Added ability to save and load BIP animation files via both dialog and quick modes.
// Version 2.4 10/05/20;
//						- Added a new baking option for keying a Biped head in order to track an active Look At Target if assigned.
// Version 2.3 28/06/17;
//						- Fixed a bug where paste posture would crash due to missing posture in the collection. It now filters the selection for
//						    biped only parts during the copy posture phase, so that the collection is created properly for pasting.
//						- UI now remembers it's last location, even after Max restart.
// Version 2.2 23/11/16;
//						- Added the ability to copy and paste (with optional opposite) a posture based on a selection of biped parts. The copied
//						  posture is saved as a file for later retrieval. You can also quickly select the copied biped parts if available.
// Version 2.1 19/11/16;
//						- Added a new rollout called Quick Posture, with options to mirror limbs, and straighten or untwist spine links.
// Version 2.0 06/02/16;
//						- Added a new button "Key Selected Biped Parts". Adds keys to all selected biped parts at the current slider time.
// Version 1.9 08/11/15;
//						- Bake Animation Keys now refreshes the UI in the case of the root being a <deleted scene node>.
//						- Trajectories Toggle now force switches to the motion panel before toggling the trajectories mode.
//						- Baking up to next and last modes now allows any part of a limb to be selected, not just the hands or feet.
//						- Added options to enable a single pose to be grabbed then applied without using the Biped collections panel. It saves the pose to a file for storage.
//						- Delete Frame Keys can now remove keys on all Biped parts at the current slider time, by holding Shift or Control when clicking the button.
//						- Added a select Both Hands & Feet button to the Selection/Display/Keys rollout.
//						- Updated PDF guide.
// Version 1.8 06/11/15;
//						- Fixed global variable usage which caused the functions key all and delete all keys to not work (finally).
//						- Changed interface references of IK to be Body/Object space instead, to more accurately reflect their purpose.
//						- Added a Delete Frame Keys button, which deletes all keys on selected Biped parts based on the slider time.
//						- Added new button images, tooltips and rearranged some UI elements.
//						- Improved code for baking animation keys.
//						- Added a PDF guide and button link to the UI (in the About rollout).
//						- The previous dialog version of this tool is no longer supported. "Compact" reference removed.
// Version 1.7 01/07/15;
//						- Converted the dialog interface into a more compact floater with collaspable rollouts.
//						- Added functionality to enable user specified snapshots scene layer.
// Version 1.6 17/12/14;
//						- Fixed global variable usage which caused the functions key all and delete all keys to not work.
// Version 1.5 09/12/14;
//						- While using the Shift or Control method for setting TCB linear or smooth values across the time range, if the
//							Escape key is pressed the operation will now stop.
// Version 1.4 27/10/14;
//						- Email change for help contact.
// Version 1.3 13/10/14;
//						- When selecting from the dropdown list of multiple Bipeds, you can now hold Shift to also select the Biped root
//							or hold Control to select the whole rig.
// Version 1.2 22/09/14;
//						- Fixed the snapshot functionality. Do longer selects anything which means it works for frozen objects in Max 2015.
//							Also, now handles multiple skin meshes in the Mesh Character layer. Only Skin meshes are snapshot.
// Version 1.1 17/09/14;
//						- Added a dropdown list to enable selection of a Biped for the panel to operate on in the case of multiple Bipeds.
// Version 1.0 15/09/14;
//						- First public release.
//
//
// This script is not copyrighted, please feel free to modify or distribute.
//
**************************************************************************************************************************/


global g_LBTools_BH

struct s_LBTools_bipedHelper
(

/*==================================================================================
// Structure variables
====================================================================================*/

BH_pos,
bhINIfile = getDir #plugcfg + "\LBTools_BipedHelper.ini",
quickBipFile = getDir #plugcfg + "\LBTools_BipedHelper_QuickSave.bip",
f_openBipedUI,
bipedUI_Floater,
AlignBake_Rollout,
SelectDisplay_Rollout,
leftHandPivots_Rollout,
rightHandPivots_Rollout,
leftFootPivots_Rollout,
rightFootPivots_Rollout,
Keys_Rollout,
BipFiles_Rollout,
PosePosture_Rollout,
LayersRetargeting_Rollout,
Layers_Rollout,
LimbSpaces_Rollout,
Reference_Rollout,
About_Rollout,
a_storedSpaces = #(),
a_bips = #(),
bRoot, bPelvis, bSpine, bHead,
bClavicleL, bUpperarmL, bForearmL, bHandL,
bClavicleR, bUpperarmR, bForearmR, bHandR,
bThighL, bCalfL, bFootL,
bThighR, bCalfR, bFootR,
bTail,
bProp1, bProp2, bProp3,
a_armLeft = #(),
a_armRight = #(),
a_legLeft = #(),
a_legRight = #(),
a_fingersLeft = #(),
a_fingersRight = #(),
a_toesLeft = #(),
a_toesRight = #(),
a_bipedHandsFeet = #(),
a_bipedLimbs = #(),
a_bipedAllParts = #(),
TRstart = animationRange.start,
TRend = animationRange.end,
TRstep = 1,
TRstate = 1,
retargetingOption = 1,
leftHandUIoffset = [-95,120],
rightHandUIoffset = [215,120],
leftFootUIoffset = [-95,280],
rightFootUIoffset = [215,280],
layersRetargetingUIoffset = [220,300],

/*==================================================================================
// Structure functions
/*==================================================================================

f_findSkinModifier
f_getAllChildren
f_findHierarchyRoot
f_findHierarchy
f_missingBipedError
f_filterBipedParts
f_findAllParts
f_findFingerRoot
f_getFingerChildren
f_findAllFingers
f_findAllToes
f_initBipedParts
f_collectBipRoots
f_processTCBvalues
f_evaluateTCBprocess
f_storeSpaces
f_loadSpaces
f_limbBody
f_limbObject
f_changePivot
f_getPivotIndex
f_isValidBiped
f_bipedBakeSlidingKeys
f_bipedBakeSlidingKeyNext
f_bipedBakeSlidingKeyLast
f_bipedBakeRoot
f_bipedBakeHead
f_bipedHelperUI

====================================================================================*/

fn f_findSkinModifier obj =
(
	local skinFound = false
	for i = 1 to obj.modifiers.count do
	(
		if classof obj.modifiers[i] == Skin then
		(
			skinFound = true
			exit
		)
	)
	skinFound
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_getAllChildren obj &a_nodes:#() = --// Recursive function
(
	for c in obj.children do
	(
		append a_nodes c
		f_getAllChildren c a_nodes:a_nodes
	)
	a_nodes
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_findHierarchyRoot obj = --// Recursive function
(
	if obj != undefined then
	(
		if isValidnode obj.parent then
		(
			f_findHierarchyRoot obj.parent
		)
		else
		(
			obj
		)
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_findHierarchy obj =
(
	if obj != undefined then 
	(
		local parentNode = g_LBTools_BH.f_findHierarchyRoot obj
		g_LBTools_BH.f_getAllChildren parentNode a_nodes:#(parentNode)
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_missingBipedError =
(
	messageBox "Error: A previously picked Biped is missing from the scene - UI updated." title:"Biped Helper"
	setIniSetting g_LBTools_BH.bhINIfile "BipedPanel" "CurrentBip" "1"
	g_LBTools_BH.f_openBipedUI()
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_filterBipedParts a_objs =
(
	local a_bipedObjs = #()
	for obj in a_objs where (classof obj.baseObject == Biped_Object) do
	(
		append a_bipedObjs obj
	)
	a_bipedObjs
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_findAllParts =
(
	g_LBTools_BH.a_bipedAllParts = #()
	local a_bipedObjs = g_LBTools_BH.f_findHierarchy g_LBTools_BH.bRoot
	a_bipedObjs = g_LBTools_BH.f_filterBipedParts a_bipedObjs
	for obj in a_bipedObjs do
	(
		if NOT (matchpattern obj.name pattern:"*Footsteps") AND NOT (matchpattern obj.name pattern:"*Twist*") AND NOT (matchpattern obj.name pattern:"*Nub") do
		(
			append g_LBTools_BH.a_bipedAllParts obj
		)
	)
	g_LBTools_BH.a_bipedAllParts
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_findFingerRoot obj = --// Recursive function
(
	if obj != undefined then
	(
		if isValidnode obj.parent AND obj.parent != g_LBTools_BH.bHandL AND obj.parent != g_LBTools_BH.bHandR then
		(
			f_findFingerRoot obj.parent
		)
		else
		(
			obj
		)
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_getFingerChildren obj &a_nodes:#() = --// Recursive function
(
	if obj != undefined then
	(
		for c in obj.children where classof c == Biped_Object do
		(
			append a_nodes c
			f_getFingerChildren c a_nodes:a_nodes
		)
	)
	a_nodes
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_findAllFingers =
(
	g_LBTools_BH.a_fingersLeft = #()
	g_LBTools_BH.a_fingersRight = #()
	local handChildren = #()
	g_LBTools_BH.f_getAllChildren g_LBTools_BH.bHandL a_nodes:handChildren
	for obj in handChildren where classof obj == Biped_Object do
	(
		append g_LBTools_BH.a_fingersLeft obj
	)
	handChildren = #()
	g_LBTools_BH.f_getAllChildren g_LBTools_BH.bHandR a_nodes:handChildren
	for obj in handChildren where classof obj == Biped_Object do
	(
		append g_LBTools_BH.a_fingersRight obj
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_findAllToes =
(
	g_LBTools_BH.a_toesLeft = #()
	g_LBTools_BH.a_toesRight = #()
	local footChildren = #()
	g_LBTools_BH.f_getAllChildren g_LBTools_BH.bFootL a_nodes:footChildren
	for obj in footChildren where classof obj == Biped_Object do
	(
		append g_LBTools_BH.a_toesLeft obj
	)
	footChildren = #()
	g_LBTools_BH.f_getAllChildren g_LBTools_BH.bFootR a_nodes:footChildren
	for obj in footChildren where classof obj == Biped_Object do
	(
		append g_LBTools_BH.a_toesRight obj
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_initBipedParts =
(
	if isValidNode g_LBTools_BH.bRoot then
	(
		g_LBTools_BH.bPelvis = biped.getNode g_LBTools_BH.bRoot #Pelvis
		g_LBTools_BH.bSpine = biped.getNode g_LBTools_BH.bRoot #Spine link:2
		g_LBTools_BH.bHead = biped.getNode g_LBTools_BH.bRoot #Head
		
		g_LBTools_BH.bClavicleL = biped.getNode g_LBTools_BH.bRoot #lArm link:1
		g_LBTools_BH.bUpperarmL = biped.getNode g_LBTools_BH.bRoot #lArm link:2
		g_LBTools_BH.bForearmL = biped.getNode g_LBTools_BH.bRoot #lArm link:3
		g_LBTools_BH.bHandL = biped.getNode g_LBTools_BH.bRoot #lArm link:4
		
		g_LBTools_BH.bClavicleR = biped.getNode g_LBTools_BH.bRoot #rArm link:1
		g_LBTools_BH.bUpperarmR = biped.getNode g_LBTools_BH.bRoot #rArm link:2
		g_LBTools_BH.bForearmR = biped.getNode g_LBTools_BH.bRoot #rArm link:3
		g_LBTools_BH.bHandR = biped.getNode g_LBTools_BH.bRoot #rArm link:4
		
		g_LBTools_BH.bThighL = biped.getNode g_LBTools_BH.bRoot #lLeg link:1
		g_LBTools_BH.bCalfL = biped.getNode g_LBTools_BH.bRoot #lLeg link:2
		g_LBTools_BH.bFootL = biped.getNode g_LBTools_BH.bRoot #lLeg link:3
		
		g_LBTools_BH.bThighR = biped.getNode g_LBTools_BH.bRoot #rLeg link:1
		g_LBTools_BH.bCalfR = biped.getNode g_LBTools_BH.bRoot #rLeg link:2
		g_LBTools_BH.bFootR = biped.getNode g_LBTools_BH.bRoot #rLeg link:3
		
		g_LBTools_BH.bTail = biped.getNode g_LBTools_BH.bRoot  #tail
		
		g_LBTools_BH.bProp1 = biped.getNode g_LBTools_BH.bRoot #prop1
		g_LBTools_BH.bProp2 = biped.getNode g_LBTools_BH.bRoot #prop2
		g_LBTools_BH.bProp3 = biped.getNode g_LBTools_BH.bRoot #prop3
		
		g_LBTools_BH.a_armLeft = #(g_LBTools_BH.bClavicleL, g_LBTools_BH.bUpperarmL, g_LBTools_BH.bForearmL, g_LBTools_BH.bHandL)
		g_LBTools_BH.a_armRight = #(g_LBTools_BH.bClavicleR, g_LBTools_BH.bUpperarmR, g_LBTools_BH.bForearmR, g_LBTools_BH.bHandR)
		g_LBTools_BH.a_legLeft = #(g_LBTools_BH.bThighL, g_LBTools_BH.bCalfL, g_LBTools_BH.bFootL)
		g_LBTools_BH.a_legRight = #(g_LBTools_BH.bThighR, g_LBTools_BH.bCalfR, g_LBTools_BH.bFootR)
		
		g_LBTools_BH.a_bipedLimbs = #(g_LBTools_BH.bClavicleL, g_LBTools_BH.bUpperarmL, g_LBTools_BH.bForearmL, g_LBTools_BH.bHandL,
														g_LBTools_BH.bClavicleR, g_LBTools_BH.bUpperarmR, g_LBTools_BH.bForearmR, g_LBTools_BH.bHandR,
														g_LBTools_BH.bThighL, g_LBTools_BH.bCalfL, g_LBTools_BH.bFootL,
														g_LBTools_BH.bThighR, g_LBTools_BH.bCalfR, g_LBTools_BH.bFootR
													)
		
		g_LBTools_BH.f_findAllFingers()
		g_LBTools_BH.f_findAllToes()
		g_LBTools_BH.a_bipedHandsFeet = #(g_LBTools_BH.bHandL, g_LBTools_BH.bHandR, g_LBTools_BH.bFootL, g_LBTools_BH.bFootR)
		g_LBTools_BH.f_findAllParts()
	)
	else
	(
		g_LBTools_BH.f_missingBipedError()
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_collectBipRoots =
(
	for b in geometry where (classof b.controller == Vertical_Horizontal_Turn) do
	(
		append g_LBTools_BH.a_bips b
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

mapped fn f_processTCBvalues obj val atTime:sliderTime =
(
	try
	(
		if obj == g_LBTools_BH.bRoot then
		(
			(biped.getKey obj.controller[1].controller (getkeyindex obj.controller[1].controller atTime)).continuity = val
			(biped.getKey obj.controller[2].controller (getkeyindex obj.controller[2].controller atTime)).continuity = val
			(biped.getKey obj.controller[3].controller (getkeyindex obj.controller[3].controller atTime)).continuity = val
		)
		else
		(
			(biped.getKey obj.controller (getkeyindex obj.controller atTime)).continuity = val
		)
	)catch(g_LBTools_BH.f_initBipedParts())
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_evaluateTCBprocess val =
(
	undo "TCBProcess" on
	(
		if selection.count > 0 then
		(
			if keyboard.shiftPressed OR keyboard.controlPressed then
			(
				for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
				(
					g_LBTools_BH.f_processTCBvalues (selection as array) val atTIme:t
				)
			)
			else
			(
				g_LBTools_BH.f_processTCBvalues (selection as array) val
			)
		)
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_storeSpaces =
(
	local a_limb_nodes = #(g_LBTools_BH.bHandL,g_LBTools_BH.bHandR,g_LBTools_BH.bFootL,g_LBTools_BH.bFootR)
	local a_limb_ctrls = #(g_LBTools_BH.bHandL.controller,g_LBTools_BH.bHandR.controller,g_LBTools_BH.bFootL.controller,g_LBTools_BH.bFootR.controller)
	local a_allKeys = #()
	local a_spaces = #(#(), #(), #(), #())

	select a_limb_nodes
	for n = 1 to a_limb_nodes.count while not keyboard.escPressed do
	(
		local keyCount = numKeys a_limb_ctrls[n]
		for i = 1 to keyCount while not keyboard.escPressed do
		(
			a_allKeys[i] = biped.getkey a_limb_ctrls[n] i
			if a_allKeys[i].ikspace == 1 then --// object space
			(
				if a_allKeys[i].ikJoinedPivot then a_spaces[n][i] = 2 --// planted key
				else a_spaces[n][i] = 1 --// sliding key
			)
			else --// body space
			(
				a_spaces[n][i] = 0 --// free key
			)
		)
		for j = 1 to keyCount while not keyboard.escPressed do
		(
			t = getKeyTime a_limb_ctrls[n] j
			sliderTime = t
			biped.setFreeKey a_limb_nodes[n]
		)
	)
	sliderTime = animationRange.start
	a_spaces
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_loadSpaces a_spaces =
(
	local a_limb_nodes = #(g_LBTools_BH.bHandL,g_LBTools_BH.bHandR,g_LBTools_BH.bFootL,g_LBTools_BH.bFootR)
	local a_limb_ctrls = #(g_LBTools_BH.bHandL.controller,g_LBTools_BH.bHandR.controller,g_LBTools_BH.bFootL.controller,g_LBTools_BH.bFootR.controller)
	
	select a_limb_nodes
	for n = 1 to a_limb_ctrls.count while not keyboard.escPressed do
	(
		for i = 1 to a_spaces[n].count while not keyboard.escPressed do
		(
			t = getKeyTime a_limb_ctrls[n] i
			sliderTime = t
			if a_spaces[n][i] == 2 then
			(
				biped.setPlantedKey a_limb_nodes[n]
			)
			if a_spaces[n][i] == 1 then
			(
				biped.setSlidingKey a_limb_nodes[n]
			)
		)
	)
	sliderTime = animationRange.start
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_limbBody =
(
	if selection.count > 0 do
	(
		for obj in (selection as array) do
		(
			try
			(			
				local keyCount = numKeys obj.controller
				for i = 1 to keyCount while not keyboard.escPressed do
				(
					t = getKeyTime obj.controller i
					sliderTime = t
					biped.setFreeKey obj
				)
				sliderTime = animationRange.start
			) catch()
		)
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_limbObject =
(
	if selection.count > 0 do
	(
		for obj in (selection as array) do
		(
			try
			(
				local keyCount = numKeys obj.controller
				for i = 1 to keyCount while not keyboard.escPressed do
				(
					t = getKeyTime obj.controller i
					sliderTime = t
					biped.setSlidingKey obj
				)
				sliderTime = animationRange.start
			) catch()
		)
	)
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_changePivot ctrl index =
(
	if ctrl.keys.count > 0 then
	(
		local keyIndex = getKeyIndex ctrl slidertime
		if keyIndex > 0 then
		(
			local key = biped.getKey ctrl keyIndex
			if key.ikSpace > 0 then
			(
				key.ikPivotIndex = index
				return true
			)
			else
			(
				messagebox "Error: must be a planted or sliding key (i.e. Object space)." title:"Pivots Selection"
			)
		)
		else
		(
			messagebox "Error: no key at slider time." title:"Pivots Selection"
		)
	)
	else
	(
		messagebox "Error: no keys found.\n\nSet a planted or sliding key." title:"Pivots Selection"
	)
	
	return false
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_getPivotIndex ctrl =
(
	if ctrl.keys.count > 0 then
	(
		local keyIndex = getKeyIndex ctrl slidertime
		if keyIndex > 0 then
		(
			local key = biped.getKey ctrl keyIndex
			if key.ikSpace > 0 then
			(
				return (biped.getKey ctrl keyIndex).ikPivotIndex
			)
		)
	)
	return 0	
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_isValidBiped =
(
	if isValidNode g_LBTools_BH.bRoot then
	(
		return true
	)
	return false
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_bipedBakeSlidingKeys a_limbs =
(
	undo "BipedBakeSlidingKeys" on
	(
		for obj in a_limbs while not keyboard.escPressed do 
		(
			select obj
			indexInitial = numKeys obj.controller
			if indexInitial > 1 then
			(
				indexAdded = 0
				for i = 1 to indexInitial while not keyboard.escPressed do
				(
					i = i + indexAdded
					local indexCount = indexInitial + indexAdded
					local currentKey = biped.getKey obj.controller i
					if i < indexCount then
					(
						nextKey = biped.getKey obj.controller (i+1)
					)
					if nextKey != undefined then
					(
						if currentKey.ikSpace == 1 and nextKey.ikSpace == 1 then --// Both keys are Object space (sliding)
						(
							t = getKeyTime obj.controller i
							if t < animationRange.start then
							(
								t = animationRange.start - 1
								sliderTime = t
							)
							if i < (indexCount + indexAdded) then --// Set sliding keys until the next sliding key is reached
							(
								nextKeyTime = getKeyTime obj.controller (i+1)
								if nextKeyTime > animationRange.end then
								(
									nextKeyTime = animationRange.end + 1
								)
								for h = (t + 1) to (nextKeyTime - 1) while not keyboard.escPressed do
								(
									sliderTime = h
									biped.SetSlidingKey obj
									indexAdded = indexAdded + 1
								)
							)
						)
					)
				)
				bRoot.controller.osObject = undefined
			)
			else
			(
				str_error = "Error: There must be at least two sliding keys on the limb > " + obj.name
				format (str_error + "\n")
				messageBox str_error title:"Biped Bake"
			)
		)
	)
	sliderTime = animationRange.start
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_bipedBakeSlidingKeyNext =
(
	local obj = selection[1]

	undo "BipedBakeSlidingKeyNext" on
	(
		while (getKeyIndex obj.controller sliderTime) == 0 do
		(
			if sliderTime == animationRange.end then
			(
				exit()
			)
			else
			(
				sliderTime += 1
			)
		)
		if sliderTime < animationRange.end then
		(
			local index = getKeyIndex obj.controller sliderTime
			local initialKey = biped.getKey obj.controller index
			if index < (numKeys obj.controller) then --// Needs at least two keys to work
			(
				local nextKey = biped.getKey obj.controller (index + 1)
				local nextKeyTime = getKeyTime obj.controller (index + 1)
				if nextKeyTime <= animationRange.end then
				(
					if initialKey.ikSpace == 1 AND nextKey.ikSpace == 1 then
					(
						local t = getKeyTime obj.controller index
						sliderTime = t
						for f = (t + 1) to (nextKeyTime - 1) while NOT keyboard.escPressed do
						(
							sliderTime = f
							biped.SetSlidingKey obj
						)
						sliderTime += 1
					)
					else
					(
						sliderTime += 1
					)
				)
				else
				(
					messageBox "Error: The next sliding key must be within the visible time range." title:"Biped Bake"
				)
			)
			else
			(
				local str_error = "Error: There must be at least two sliding keys on the limb " + obj.name + " from the slider time."
				format (str_error + "\n")
				messageBox str_error title:"Biped Bake"
			)
		)
	)
	sliderTime = animationRange.start
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_bipedBakeSlidingKeyLast =
(
	local bool_done = false
	local obj = selection[1]
	
	undo "BipedBakeSlidingKeyLast" on
	(
		while bool_done == false do
		(
			while (getKeyIndex obj.controller sliderTime) == 0 do
			(
				if sliderTime == animationRange.end then
				(
					exit()
				)
				else
				(
					sliderTime += 1
				)
			)
			if sliderTime < animationRange.end then
			(
				local index = getKeyIndex obj.controller sliderTime
				local currentKey = biped.getKey obj.controller index
				if index < (numKeys obj.controller) then --// Needs at least two keys to work
				(
					local nextKey = biped.getKey obj.controller (index + 1)
					local nextKeyTime = getKeyTime obj.controller (index + 1)
					if nextKeyTime <= animationRange.end then
					(
						if currentKey.ikSpace == 1 and nextKey.ikSpace == 1 then
						(
							local t = getKeyTime obj.controller index
							sliderTime = t
							for f = (t + 1) to (nextKeyTime - 1) while not keyboard.escPressed do
							(
								sliderTime = f
								biped.SetSlidingKey obj
							)
							sliderTime += 1
							index = getKeyIndex obj.controller sliderTime
							currentKey = biped.getKey obj.controller index
							if index < (numKeys obj.controller) then
							(
								nextKey = biped.getKey obj.controller (index + 1)
								if nextKey.ikSpace == 0 then
								(
									bool_done = true
								)
							)
							else
							(
								exit()
							)
						)
						else
						(
							sliderTime += 1
						)
					)
					else
					(
						messageBox "Error: The next sliding key must be within the visible time range." title:"Biped Bake"
						exit()
					)
				)
				else
				(
					local str_error = "Error: There must be at least two sliding keys on the limb " + obj.name + " from the slider time."
					format (str_error + "\n")
					messageBox str_error title:"Biped Bake"
					exit()
				)
			)
			else
			(
				bool_done = true
			)
		)
	)
	sliderTime = animationRange.start
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_bipedBakeRoot =
(
	sliderTime = animationRange.start
	local bipObj = g_LBTools_BH.bRoot
	select bipObj
	linkCtrl = bipObj.parent
	local followHelper = dummy()
	followHelper.name = "FollowHelper"
	followHelper.parent = linkCtrl
	with animate on
	(
		undo "BipedBakeRoot" on
		(
			for t = animationRange.start to animationRange.end while not keyboard.escPressed do
			(
				at time t (followHelper.transform = bipObj.transform)
			)
			bipObj.parent = undefined
			if classof linkCtrl.controller == Link_Constraint then
			(
				for i = 1 to linkCtrl.controller.getNumTargets() while not keyboard.escPressed do
				(
					if linkCtrl.controller.getNode i != undefined then --// A target object is set (not linked to world space)
					(
						local currentTime = linkCtrl.controller.getFrameNo i --// Current target time
						if i < linkCtrl.controller.getNumTargets() then --// Is not the last target
						(
							local nextTime = linkCtrl.controller.getFrameNo (i+1) --// Get the next target time
							if currentTime >= animationRange.start and nextTime <= animationRange.end then
							(
								for t = currentTime to nextTime while not keyboard.escPressed do
								(
									sliderTime = t
									bipObj.transform = followHelper.transform
								)
							)
							else
							(
								for t = animationRange.start to animationRange.end while not keyboard.escPressed do
								(
									sliderTime = t
									bipObj.transform = followHelper.transform
								)
							)
						)
						if i == linkCtrl.controller.getNumTargets() and sliderTime != animationRange.end then --// Is the last target
						(
							for t = currentTime to animationRange.end while not keyboard.escPressed do
							(
								sliderTime = t
								bipObj.transform = followHelper.transform
							)					
						)
					)
				)
			)
			else
			(
				for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
				(
					sliderTime = t
					bipObj.transform = followHelper.transform
				)
			)
		)
	)
	delete followHelper
	sliderTime = animationRange.start
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_bipedBakeHead =
(
	local pt = Point()
	with animate on
	(
		undo "BipedBakeHead" on
		(
			for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
			(
				at time t (pt.transform = g_LBTools_BH.bHead.transform)
			)

			g_LBTools_BH.bRoot.controller.headObject = undefined
			
			for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
			(
				at time t (g_LBTools_BH.bHead.transform = pt.transform)
			)
		)
	)
	delete pt
	sliderTime = animationRange.start
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_bipedHelperUI =
(	
	rollout SelectDisplay_Rollout "Selection / Display / Time"
	(
		dropDownList ddl_bips "" width:150 across:2 align:#left tooltip:"Choose a Biped root node from the list for the panel to operate on. Hold SHIFT to select the Biped root or CONTROL to select the whole rig."
		button btn_refreshUI "R" width:20 height:20 align:#right offset:[8,0] tooltip:"Refresh the list by searching for all Biped roots in the scene."
		group "Selection"
		(
			label lbl_txt1 "Add: Shift/Control - Sub: Alt" align:#center
			button btn_selectAll "Select All" align:#center width:160 height:20 tooltip:"Select all Biped parts."
			button btn_selectHead "Select Head" align:#center width:160 height:20 tooltip:"Select the head control."
			button btn_selectSpine "Select Spine" align:#center width:160 height:20 tooltip:"Select the second spine link."
			button btn_selectRoot "Select Root" align:#center width:160 height:20 tooltip:"Select the root control (Bip01 or Bip001)."
			button btn_selectFingersLeft "L Fingers" align:#left width:78 height:20 across:2 offset:[3,0] tooltip:"Select all the left hand fingers."
			button btn_selectFingersRight "R Fingers" align:#right width:78 height:20 offset:[-3,0] tooltip:"Select all the right hand fingers."
			checkbutton chkbtn_pivotsLHand "P" across:4 width:14 height:20 align:#left offset:[3,0] tooltip:"Open pivots selection for the left hand."
			button btn_selectLHand "Left Hand" width:63 height:20 offset:[-11,0] tooltip:"Select the left hand control."
			button btn_selectRHand "Right Hand" width:63 height:20 offset:[17,0] tooltip:"Select the right hand control."
			checkbutton chkbtn_pivotsRHand "P" width:14 height:20 offset:[14,0] tooltip:"Open pivots selection for the right hand."
			checkbutton chkbtn_pivotsLFoot "P" across:4 width:14 height:20 align:#left offset:[3,0] tooltip:"Open pivots selection for the left foot."
			button btn_selectLFoot "Left Foot" width:63 height:20 offset:[-11,0] tooltip:"Select the left foot control."
			button btn_selectRFoot "Right Foot" width:63 height:20 offset:[17,0] tooltip:"Select the right foot control."
			checkbutton chkbtn_pivotsRFoot "P" width:14 height:20 offset:[14,0] tooltip:"Open pivots selection for the right foot."
			button btn_selectHandsFeet "Both Hands & Feet" align:#center width:160 height:20 tooltip:"Select all hand and foot controls."
			button btn_selectProp1 "Prop 1" align:#left width:50 height:20 across:3 offset:[3,0] tooltip:"Select prop 1 control if available."
			button btn_selectProp2 "Prop 2" align:#center width:50 height:20 tooltip:"Select prop 2 control if available."
			button btn_selectProp3 "Prop 3" align:#right width:50 height:20 offset:[-3,0] tooltip:"Select prop 3 control if available."
		)
		group "Display"
		(
			button btn_hide "Hide Biped" align:#left width:78 height:20 across:2 offset:[3,0] tooltip:"Hide all Biped parts."
			button btn_unhide "Unhide Biped" align:#right width:78 height:20 offset:[-3,0] tooltip:"Unhide Biped parts except for end nubs and the footsteps node."
			button btn_trajToggle "Trajectories Toggle" align:#center width:160 height:20 tooltip:"Toggle trajectories display for selected parts."
		)
		group "Time Range"
		(
			radiobuttons rdobtns_timeRange labels:#("Active Range", "Set Range") default:1 columns:2 align:#center offset:[5,0] offsets:#([0,0],[5,0])
			spinner spn_TRstartTime "S" range:[0,99999, g_LBTools_BH.TRstart] type:#integer fieldwidth:40 enabled:false align:#left across:3 tooltip:"Animation range START for baking."
			spinner spn_TRendTime "E" range:[1,99999, g_LBTools_BH.TRend] type:#integer fieldwidth:40 align:#center enabled:false offset:[7,0] tooltip:"Animation range END for baking."
			spinner spn_TRstepSize "BY" range:[1,99, g_LBTools_BH.TRstep] type:#integer fieldwidth:20 align:#right tooltip:"STEP SIZE for baking."
			button btn_setTRstart "" width:55 height:10 align:#left offset:[0,-3] across:2 enabled:false tooltip:"Set the animation START time to the current slider time."
			button btn_setTRend "" width:55 height:10 align:#left offset:[-22,-3] enabled:false tooltip:"Set the animation END time to the current slider time."
		)	
		
		fn f_closeAllDialogs =
		(
			try(destroyDialog g_LBTools_BH.leftHandPivots_Rollout)catch()
			try(destroyDialog g_LBTools_BH.rightHandPivots_Rollout)catch()
			try(destroyDialog g_LBTools_BH.leftFootPivots_Rollout)catch()
			try(destroyDialog g_LBTools_BH.rightFootPivots_Rollout)catch()
			try(destroyDialog g_LBTools_BH.LayersRetargeting_Rollout)catch()
		)

		fn f_selectBipedParts a_parts =
		(
			setCommandPanelTaskMode #motion
			try
			(
				if keyboard.shiftPressed OR keyboard.controlPressed then
				(
					selectmore a_parts
				)
				else
				(
					if keyboard.altPressed then
					(
						deselect a_parts
					)
					else
					(
						select a_parts
					)
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on SelectDisplay_Rollout open do
		(
			setCommandPanelTaskMode #motion
			local a_bipNames = #()
			for b in g_LBTools_BH.a_bips do
			(
				append a_bipNames b.name
			)
			ddl_bips.items = a_bipNames
			ddl_bips.selection = execute (getIniSetting g_LBTools_BH.bhINIfile "BipedPanel" "CurrentBip")
			if ddl_bips.selection > 0 then
			(
				g_LBTools_BH.bRoot = g_LBTools_BH.a_bips[ddl_bips.selection]
				g_LBTools_BH.f_initBipedParts()
				ddl_bips.height = (ddl_bips.items.count + 1) * 100
			)
			else
			(
				g_LBTools_BH.f_missingBipedError()					
			)
			spn_TRstartTime.value = g_LBTools_BH.TRstart = animationRange.start
			spn_TRendTime.value = g_LBTools_BH.TRend = animationRange.end
			spn_TRstepSize.value = g_LBTools_BH.TRstep = 1
		)

		on SelectDisplay_Rollout close do
		(
			f_closeAllDialogs()
		)
		
		on SelectDisplay_Rollout moved pos do
		(
			g_LBTools_BH.BH_pos = pos
			setIniSetting g_LBTools_BH.bhINIfile "Floater" "Position" (g_LBTools_BH.BH_pos as string)

			--// update any external dialog positions (if open)
			SetDialogPos g_LBTools_BH.leftHandPivots_Rollout (g_LBTools_BH.BH_pos + g_LBTools_BH.leftHandUIoffset)
			SetDialogPos g_LBTools_BH.rightHandPivots_Rollout (g_LBTools_BH.BH_pos + g_LBTools_BH.rightHandUIoffset)
			SetDialogPos g_LBTools_BH.leftFootPivots_Rollout (g_LBTools_BH.BH_pos + g_LBTools_BH.leftFootUIoffset)
			SetDialogPos g_LBTools_BH.rightFootPivots_Rollout (g_LBTools_BH.BH_pos + g_LBTools_BH.rightFootUIoffset)			
			SetDialogPos g_LBTools_BH.LayersRetargeting_Rollout (g_LBTools_BH.BH_pos + g_LBTools_BH.layersRetargetingUIoffset)
		)
		
		on ddl_bips selected val do
		(
			if isValidNode g_LBTools_BH.a_bips[ddl_bips.selection] then
			(
				if ddl_bips.selected == g_LBTools_BH.a_bips[ddl_bips.selection].name then
				(
					g_LBTools_BH.bRoot = g_LBTools_BH.a_bips[ddl_bips.selection]
					g_LBTools_BH.f_initBipedParts()
					setIniSetting g_LBTools_BH.bhINIfile "BipedPanel" "CurrentBip" (ddl_bips.selection as string)
					setCommandPanelTaskMode #motion
					if keyboard.shiftPressed do
					(
						try(select g_LBTools_BH.bRoot)catch(g_LBTools_BH.f_initBipedParts())
					)
					if keyboard.controlPressed do
					(
						try(select g_LBTools_BH.a_bipedAllParts)catch(g_LBTools_BH.f_initBipedParts())
					)
					f_closeAllDialogs()
					g_LBTools_BH.Layers_Rollout.f_updateLayers()
				)
				else
				(
					g_LBTools_BH.f_missingBipedError()
				)
			)
			else
			(
				g_LBTools_BH.f_missingBipedError()
			)
		)
		
		on btn_refreshUI pressed do
		(
			g_LBTools_BH.f_openBipedUI()
		)

		on btn_selectAll pressed do
		(
			setCommandPanelTaskMode #motion
			try
			(
				select g_LBTools_BH.a_bipedAllParts
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_selectHead pressed do
		(
			f_selectBipedParts g_LBTools_BH.bHead
		)
		
		on btn_selectSpine pressed do
		(
			f_selectBipedParts g_LBTools_BH.bSpine
		)
		
		on btn_selectRoot pressed do
		(
			f_selectBipedParts g_LBTools_BH.bRoot
		)
		
		on btn_selectFingersLeft pressed do
		(
			f_selectBipedParts g_LBTools_BH.a_fingersLeft
		)
		
		on btn_selectFingersRight pressed do
		(
			f_selectBipedParts g_LBTools_BH.a_fingersRight
		)
		
		on btn_selectLFoot pressed do
		(
			f_selectBipedParts g_LBTools_BH.bFootL
		)
	
		on chkbtn_pivotsLFoot changed state do
		(
			try
			(
				if state then
				(
					try(destroyDialog g_LBTools_BH.leftFootPivots_Rollout)catch()
					createDialog g_LBTools_BH.leftFootPivots_Rollout width:90 height:70 pos:(g_LBTools_BH.BH_pos + g_LBTools_BH.leftFootUIoffset) style:#(#style_toolwindow, #style_sysmenu)
				)
				else
				(
					try(destroyDialog g_LBTools_BH.leftFootPivots_Rollout)catch()
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_selectRFoot pressed do
		(
			f_selectBipedParts g_LBTools_BH.bFootR
		)
		
		on chkbtn_pivotsRFoot changed state do
		(
			try
			(
				if state then
				(
					try(destroyDialog g_LBTools_BH.rightFootPivots_Rollout)catch()
					createDialog g_LBTools_BH.rightFootPivots_Rollout width:90 height:70 pos:(g_LBTools_BH.BH_pos + g_LBTools_BH.rightFootUIoffset) style:#(#style_toolwindow, #style_sysmenu)
				)
				else
				(
					try(destroyDialog g_LBTools_BH.rightFootPivots_Rollout)catch()
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_selectLHand pressed do
		(
			f_selectBipedParts g_LBTools_BH.bHandL
		)
	
		on chkbtn_pivotsLHand changed state do
		(
			try
			(
				if state then
				(
					try(destroyDialog g_LBTools_BH.leftHandPivots_Rollout)catch()
					createDialog g_LBTools_BH.leftHandPivots_Rollout width:90 height:120 pos:(g_LBTools_BH.BH_pos + g_LBTools_BH.leftHandUIoffset) style:#(#style_toolwindow, #style_sysmenu)
				)
				else
				(
					try(destroyDialog g_LBTools_BH.leftHandPivots_Rollout)catch()
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_selectRHand pressed do
		(
			f_selectBipedParts g_LBTools_BH.bHandR
		)
		
		on chkbtn_pivotsRHand changed state do
		(
			try
			(
				if state then
				(
					try(destroyDialog g_LBTools_BH.rightHandPivots_Rollout)catch()
					createDialog g_LBTools_BH.rightHandPivots_Rollout width:90 height:120 pos:(g_LBTools_BH.BH_pos + g_LBTools_BH.rightHandUIoffset) style:#(#style_toolwindow, #style_sysmenu)
				)
				else
				(
					try(destroyDialog g_LBTools_BH.rightHandPivots_Rollout)catch()
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_selectHandsFeet pressed do
		(
			f_selectBipedParts g_LBTools_BH.a_bipedHandsFeet
		)
		
		on btn_selectProp1 pressed do
		(
			f_selectBipedParts g_LBTools_BH.bProp1
		)
		
		on btn_selectProp2 pressed do
		(
			f_selectBipedParts g_LBTools_BH.bProp2
		)
		
		on btn_selectProp3 pressed do
		(
			f_selectBipedParts g_LBTools_BH.bProp3
		)
		
		on btn_hide pressed do
		(
			try
			(
				local a_bipedObjs = g_LBTools_BH.f_findHierarchy g_LBTools_BH.bRoot
				a_bipedObjs = g_LBTools_BH.f_filterBipedParts a_bipedObjs
				for h in helpers do
				(
					if matchpattern h.name pattern:"*Nub" do
					(
						append a_bipedObjs h
					)
				)
				hide a_bipedObjs
			)catch(g_LBTools_BH.f_initBipedParts())
		)

		on btn_unhide pressed do
		(
			try
			(
				unhide g_LBTools_BH.a_bipedAllParts
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_trajToggle pressed do
		(
			try
			(
				setCommandPanelTaskMode #motion				
				local bip = g_LBTools_BH.bRoot
				if bip.controller.displaytrajectories then
				(
					bip.controller.displaytrajectories = false
				)
				else
				(
					bip.controller.displaytrajectories = true
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on rdobtns_timeRange changed state do
		(
			case rdobtns_timeRange.state of
			(
				1: (
						g_LBTools_BH.TRstart = animationRange.start
						g_LBTools_BH.TRend = animationRange.end
						spn_TRstartTime.enabled = false
						spn_TRendTime.enabled = false
						btn_setTRstart.enabled = false
						btn_setTRend.enabled = false
					)
				2: (
						g_LBTools_BH.TRstart = spn_TRstartTime.value as time
						g_LBTools_BH.TRend = spn_TRendTime.value as time
						spn_TRstartTime.enabled = true
						spn_TRendTime.enabled = true
						btn_setTRstart.enabled = true
						btn_setTRend.enabled = true
					)
			)
		)

		on spn_TRstartTime changed val do
		(
			if val >= spn_TRendTime.value then
			(
				g_LBTools_BH.TRend = spn_TRendTime.value = val + 1
			)
			g_LBTools_BH.TRstart = val as time
		)
		
		on spn_TRendTime changed val do
		(
			if val <= spn_TRstartTime.value then
			(
				g_LBTools_BH.TRstart = spn_TRstartTime.value = val - 1
			)
			g_LBTools_BH.TRend = val as time
		)
		
		on spn_TRstepSize changed val do
		(
			g_LBTools_BH.TRstep = val
		)
			
		on btn_setTRstart pressed do
		(
			spn_TRstartTime.value = sliderTime
			if sliderTime >= spn_TRendTime.value then
			(
				g_LBTools_BH.TRend = spn_TRendTime.value = sliderTime + 1
			)
			g_LBTools_BH.TRstart = sliderTime
		)
		
		on btn_setTRend pressed do
		(
			spn_TRendTime.value = sliderTime
			if sliderTime <= spn_TRstartTime.value then
			(
				g_LBTools_BH.TRstart = spn_TRstartTime.value = sliderTime - 1
			)
			g_LBTools_BH.TRend = sliderTime
		)
	) --// end rollout
	
	rollout leftHandPivots_Rollout "L-Hand"
	(
		local lrg = 16
		local sml = 12

		checkbutton chkbtn_pinkyTip "" across:4 width:sml height:sml align:#left offset:[-8,0] tooltip:"pinky tip"
		checkbutton chkbtn_ringTip "" width:sml height:sml offset:[-8,0] tooltip:"ring tip"
		checkbutton chkbtn_middleTip "" width:sml height:sml offset:[-8,0] tooltip:"middle tip"
		checkbutton chkbtn_indexTip "" width:sml height:sml offset:[-8,0] tooltip:"index tip"
		
		checkbutton chkbtn_pinky2 "" width:sml height:sml across:4 align:#left offset:[-8,0] tooltip:"pinky 2"
		checkbutton chkbtn_ring2 "" width:sml height:sml offset:[-8,0] tooltip:"ring 2"
		checkbutton chkbtn_middle2 "" width:sml height:sml offset:[-8,0] tooltip:"middle 2"
		checkbutton chkbtn_index2 "" width:sml height:sml offset:[-8,0] tooltip:"index 2"
		
		checkbutton chkbtn_pinky1 "" width:sml height:sml across:4 align:#left offset:[-8,0] tooltip:"pinky 1"
		checkbutton chkbtn_ring1 "" width:sml height:sml offset:[-8,0] tooltip:"ring 1"
		checkbutton chkbtn_middle1 "" width:sml height:sml offset:[-8,0] tooltip:"middle 1"
		checkbutton chkbtn_index1 "" width:sml height:sml offset:[-8,0] tooltip:"index 1"
		
		checkbutton chkbtn_palmFront "" width:lrg height:lrg across:3 align:#left offset:[-8,0] tooltip:"palm front"
		checkbutton chkbtn_palmMiddle "" width:lrg height:lrg offset:[-9,0] tooltip:"palm middle"
		checkbutton chkbtn_palmBack "" width:lrg height:lrg offset:[-8,0] tooltip:"palm back"
		
		checkbutton chkbtn_wristFront "" width:lrg height:lrg across:3 align:#left offset:[-8,0] tooltip:"wrist front"
		checkbutton chkbtn_wristMiddle "" width:lrg height:lrg offset:[-9,0] tooltip:"wrist middle"
		checkbutton chkbtn_wristBack "" width:lrg height:lrg offset:[-8,0] tooltip:"wrist back"
		
		checkbutton chkbtn_wristCenter "" width:lrg height:lrg align:#left offset:[14,0] tooltip:"wrist center"
		
		checkbutton chkbtn_thumbTip "" width:sml height:sml align:#right offset:[7,-92] tooltip:"thumb tip"
		checkbutton chkbtn_thumb2 "" width:sml height:sml align:#right offset:[7,0] tooltip:"thumb 2"
		checkbutton chkbtn_thumb1 "" width:sml height:sml align:#right offset:[7,0] tooltip:"thumb 1"
		
		fn f_uncheckAll =
		(
			chkbtn_wristCenter.checked = false
			chkbtn_wristFront.checked = false
			chkbtn_wristMiddle.checked = false
			chkbtn_wristBack.checked = false
			chkbtn_palmFront.checked = false
			chkbtn_palmMiddle.checked = false
			chkbtn_palmBack.checked = false
			chkbtn_thumb1.checked = false
			chkbtn_thumb2.checked = false
			chkbtn_thumbTip.checked = false
			chkbtn_index1.checked = false
			chkbtn_index2.checked = false
			chkbtn_indexTip.checked = false
			chkbtn_middle1.checked = false
			chkbtn_middle2.checked = false
			chkbtn_middleTip.checked = false
			chkbtn_ring1.checked = false
			chkbtn_ring2.checked = false
			chkbtn_ringTip.checked = false
			chkbtn_pinky1.checked = false
			chkbtn_pinky2.checked = false
			chkbtn_pinkyTip.checked = false
		)
		
		fn f_updateButton state chkbtn index =
		(
			f_uncheckAll();
			if state then
			(
				if (g_LBTools_BH.f_changePivot g_LBTools_BH.bHandL.controller index) then
				(
					chkbtn.checked = true
				)
			)
			else
			(
				if (g_LBTools_BH.f_getPivotIndex g_LBTools_BH.bHandL.controller) > 0 then
				(
					chkbtn.checked = true
				)
			)
		)
		
		on leftHandPivots_Rollout open do
		(
			case (g_LBTools_BH.f_getPivotIndex g_LBTools_BH.bHandL.controller) of
			(
				1: chkbtn_wristCenter.checked = true
				2: chkbtn_wristFront.checked = true
				3: chkbtn_wristMiddle.checked = true
				4: chkbtn_wristBack.checked = true
				5: chkbtn_palmFront.checked = true
				6: chkbtn_palmMiddle.checked = true
				7: chkbtn_palmBack.checked = true
				8: chkbtn_thumb1.checked = true
				9: chkbtn_thumb2.checked = true
				10: chkbtn_thumbTip.checked = true
				11: chkbtn_index1.checked = true
				12: chkbtn_index2.checked = true
				13: chkbtn_indexTip.checked = true
				14: chkbtn_middle1.checked = true
				15: chkbtn_middle2.checked = true
				16: chkbtn_middleTip.checked = true
				17: chkbtn_ring1.checked = true
				18: chkbtn_ring2.checked = true
				19: chkbtn_ringTip.checked = true
				20: chkbtn_pinky1.checked = true
				21: chkbtn_pinky2.checked = true
				22: chkbtn_pinkyTip.checked = true
				default: f_uncheckAll()
			)
		)
		
		on leftHandPivots_Rollout close do
		(
			g_LBTools_BH.SelectDisplay_Rollout.chkbtn_pivotsLHand.checked = false
		)
		
		on chkbtn_wristCenter changed state do (f_updateButton state chkbtn_wristCenter 1)
		on chkbtn_wristFront changed state do (f_updateButton state chkbtn_wristFront 2)
		on chkbtn_wristMiddle changed state do (f_updateButton state chkbtn_wristMiddle 3)
		on chkbtn_wristBack changed state do (f_updateButton state chkbtn_wristBack 4)
		on chkbtn_palmFront changed state do (f_updateButton state chkbtn_palmFront 5)
		on chkbtn_palmMiddle changed state do (f_updateButton state chkbtn_palmMiddle 6)
		on chkbtn_palmBack changed state do (f_updateButton state chkbtn_palmBack 7)
		on chkbtn_thumb1 changed state do (f_updateButton state chkbtn_thumb1 8)
		on chkbtn_thumb2 changed state do (f_updateButton state chkbtn_thumb2 9)
		on chkbtn_thumbTip changed state do (f_updateButton state chkbtn_thumbTip 10)
		on chkbtn_index1 changed state do (f_updateButton state chkbtn_index1 11)
		on chkbtn_index2 changed state do (f_updateButton state chkbtn_index2 12)
		on chkbtn_indexTip changed state do (f_updateButton state chkbtn_indexTip 13)
		on chkbtn_middle1 changed state do (f_updateButton state chkbtn_middle1 14)
		on chkbtn_middle2 changed state do (f_updateButton state chkbtn_middle2 15)
		on chkbtn_middleTip changed state do (f_updateButton state chkbtn_middleTip 16)
		on chkbtn_ring1 changed state do (f_updateButton state chkbtn_ring1 17)
		on chkbtn_ring2 changed state do (f_updateButton state chkbtn_ring2 18)
		on chkbtn_ringTip changed state do (f_updateButton state chkbtn_ringTip 19)
		on chkbtn_pinky1 changed state do (f_updateButton state chkbtn_pinky1 20)
		on chkbtn_pinky2 changed state do (f_updateButton state chkbtn_pinky2 21)
		on chkbtn_pinkyTip changed state do (f_updateButton state chkbtn_pinkyTip 22)
	) --// end rollout
	
	rollout rightHandPivots_Rollout "R-Hand"
	(
		local lrg = 16
		local sml = 12
		
		checkbutton chkbtn_indexTip "" across:4 width:sml height:sml align:#right offset:[8,0] tooltip:"index tip"
		checkbutton chkbtn_middleTip "" width:sml height:sml offset:[12,0] tooltip:"middle tip"
		checkbutton chkbtn_ringTip "" width:sml height:sml offset:[12,0] tooltip:"ring tip"
		checkbutton chkbtn_pinkyTip "" width:sml height:sml offset:[12,0] tooltip:"pinky tip"

		checkbutton chkbtn_index2 "" width:sml height:sml across:4 align:#right offset:[8,0] tooltip:"index 2"
		checkbutton chkbtn_middle2 "" width:sml height:sml offset:[12,0] tooltip:"middle 2"
		checkbutton chkbtn_ring2 "" width:sml height:sml offset:[12,0] tooltip:"ring 2"
		checkbutton chkbtn_pinky2 "" width:sml height:sml offset:[12,0] tooltip:"pinky 2"
		
		checkbutton chkbtn_index1 "" width:sml height:sml across:4 align:#right offset:[8,0] tooltip:"index 1"
		checkbutton chkbtn_middle1 "" width:sml height:sml offset:[12,0] tooltip:"middle 1"
		checkbutton chkbtn_ring1 "" width:sml height:sml offset:[12,0] tooltip:"ring 1"
		checkbutton chkbtn_pinky1 "" width:sml height:sml offset:[12,0] tooltip:"pinky 1"

		checkbutton chkbtn_palmFront "" width:lrg height:lrg across:3 align:#right offset:[7,0] tooltip:"palm front"
		checkbutton chkbtn_palmMiddle "" width:lrg height:lrg offset:[11,0] tooltip:"palm middle"
		checkbutton chkbtn_palmBack "" width:lrg height:lrg offset:[12,0] tooltip:"palm back"		

		checkbutton chkbtn_wristFront "" width:lrg height:lrg across:3 align:#right offset:[7,0] tooltip:"wrist front"
		checkbutton chkbtn_wristMiddle "" width:lrg height:lrg offset:[11,0] tooltip:"wrist middle"
		checkbutton chkbtn_wristBack "" width:lrg height:lrg offset:[12,0] tooltip:"wrist back"
		
		checkbutton chkbtn_wristCenter "" width:lrg height:lrg align:#right offset:[-14,0] tooltip:"wrist center"
		
		checkbutton chkbtn_thumbTip "" width:sml height:sml align:#right offset:[-58,-92] tooltip:"thumb tip"
		checkbutton chkbtn_thumb2 "" width:sml height:sml align:#right offset:[-58,0] tooltip:"thumb 2"
		checkbutton chkbtn_thumb1 "" width:sml height:sml align:#right offset:[-58,0] tooltip:"thumb 1"
			
		fn f_uncheckAll =
		(
			chkbtn_wristCenter.checked = false
			chkbtn_wristFront.checked = false
			chkbtn_wristMiddle.checked = false
			chkbtn_wristBack.checked = false
			chkbtn_palmFront.checked = false
			chkbtn_palmMiddle.checked = false
			chkbtn_palmBack.checked = false
			chkbtn_thumb1.checked = false
			chkbtn_thumb2.checked = false
			chkbtn_thumbTip.checked = false
			chkbtn_index1.checked = false
			chkbtn_index2.checked = false
			chkbtn_indexTip.checked = false
			chkbtn_middle1.checked = false
			chkbtn_middle2.checked = false
			chkbtn_middleTip.checked = false
			chkbtn_ring1.checked = false
			chkbtn_ring2.checked = false
			chkbtn_ringTip.checked = false
			chkbtn_pinky1.checked = false
			chkbtn_pinky2.checked = false
			chkbtn_pinkyTip.checked = false
		)
		
		fn f_updateButton state chkbtn index =
		(
			f_uncheckAll();
			if state then
			(
				if (g_LBTools_BH.f_changePivot g_LBTools_BH.bHandR.controller index) then
				(
					chkbtn.checked = true
				)
			)
			else
			(
				if (g_LBTools_BH.f_getPivotIndex g_LBTools_BH.bHandR.controller) > 0 then
				(
					chkbtn.checked = true
				)
			)
		)
	
		on rightHandPivots_Rollout open do
		(
			case (g_LBTools_BH.f_getPivotIndex g_LBTools_BH.bHandR.controller) of
			(
				1: chkbtn_wristCenter.checked = true
				2: chkbtn_wristFront.checked = true
				3: chkbtn_wristMiddle.checked = true
				4: chkbtn_wristBack.checked = true
				5: chkbtn_palmFront.checked = true
				6: chkbtn_palmMiddle.checked = true
				7: chkbtn_palmBack.checked = true
				8: chkbtn_thumb1.checked = true
				9: chkbtn_thumb2.checked = true
				10: chkbtn_thumbTip.checked = true
				11: chkbtn_index1.checked = true
				12: chkbtn_index2.checked = true
				13: chkbtn_indexTip.checked = true
				14: chkbtn_middle1.checked = true
				15: chkbtn_middle2.checked = true
				16: chkbtn_middleTip.checked = true
				17: chkbtn_ring1.checked = true
				18: chkbtn_ring2.checked = true
				19: chkbtn_ringTip.checked = true
				20: chkbtn_pinky1.checked = true
				21: chkbtn_pinky2.checked = true
				22: chkbtn_pinkyTip.checked = true
				default: f_uncheckAll()
			)
		)
		
		on rightHandPivots_Rollout close do
		(
			g_LBTools_BH.SelectDisplay_Rollout.chkbtn_pivotsRHand.checked = false
		)
		
		on chkbtn_wristCenter changed state do (f_updateButton state chkbtn_wristCenter 1)
		on chkbtn_wristFront changed state do (f_updateButton state chkbtn_wristFront 2)
		on chkbtn_wristMiddle changed state do (f_updateButton state chkbtn_wristMiddle 3)
		on chkbtn_wristBack changed state do (f_updateButton state chkbtn_wristBack 4)
		on chkbtn_palmFront changed state do (f_updateButton state chkbtn_palmFront 5)
		on chkbtn_palmMiddle changed state do (f_updateButton state chkbtn_palmMiddle 6)
		on chkbtn_palmBack changed state do (f_updateButton state chkbtn_palmBack 7)
		on chkbtn_thumb1 changed state do (f_updateButton state chkbtn_thumb1 8)
		on chkbtn_thumb2 changed state do (f_updateButton state chkbtn_thumb2 9)
		on chkbtn_thumbTip changed state do (f_updateButton state chkbtn_thumbTip 10)
		on chkbtn_index1 changed state do (f_updateButton state chkbtn_index1 11)
		on chkbtn_index2 changed state do (f_updateButton state chkbtn_index2 12)
		on chkbtn_indexTip changed state do (f_updateButton state chkbtn_indexTip 13)
		on chkbtn_middle1 changed state do (f_updateButton state chkbtn_middle1 14)
		on chkbtn_middle2 changed state do (f_updateButton state chkbtn_middle2 15)
		on chkbtn_middleTip changed state do (f_updateButton state chkbtn_middleTip 16)
		on chkbtn_ring1 changed state do (f_updateButton state chkbtn_ring1 17)
		on chkbtn_ring2 changed state do (f_updateButton state chkbtn_ring2 18)
		on chkbtn_ringTip changed state do (f_updateButton state chkbtn_ringTip 19)
		on chkbtn_pinky1 changed state do (f_updateButton state chkbtn_pinky1 20)
		on chkbtn_pinky2 changed state do (f_updateButton state chkbtn_pinky2 21)
		on chkbtn_pinkyTip changed state do (f_updateButton state chkbtn_pinkyTip 22)
	) --// end rollout
	
	rollout leftFootPivots_Rollout "L-Foot"
	(
		local lrg = 16

		checkbutton chkbtn_frontLeft "" across:3 width:lrg height:lrg align:#left offset:[0,0] tooltip:"front left"
		checkbutton chkbtn_frontMiddle "" width:lrg height:lrg offset:[0,0] tooltip:"front middle"
		checkbutton chkbtn_frontRight "" width:lrg height:lrg offset:[3,0] tooltip:"front right"
		checkbutton chkbtn_ankle "" width:lrg height:lrg align:#left offset:[13,0] tooltip:"ankle"
		checkbutton chkbtn_backLeft "" across:3 width:lrg height:lrg align:#left offset:[0,0] tooltip:"back left"
		checkbutton chkbtn_backMiddle "" width:lrg height:lrg offset:[0,0] tooltip:"back middle"
		checkbutton chkbtn_backRight "" width:lrg height:lrg offset:[3,0] tooltip:"back right"

		fn f_uncheckAll =
		(
			chkbtn_ankle.checked = false
			chkbtn_backLeft.checked = false
			chkbtn_backMiddle.checked = false
			chkbtn_backRight.checked = false
			chkbtn_frontLeft.checked = false
			chkbtn_frontMiddle.checked = false
			chkbtn_frontRight.checked = false
		)
		
		fn f_updateButton state chkbtn index =
		(
			f_uncheckAll();
			if state then
			(
				if (g_LBTools_BH.f_changePivot g_LBTools_BH.bFootL.controller index) then
				(
					chkbtn.checked = true
				)
			)
			else
			(
				if (g_LBTools_BH.f_getPivotIndex g_LBTools_BH.bFootL.controller) > 0 then
				(
					chkbtn.checked = true
				)
			)
		)
	
		on leftFootPivots_Rollout open do
		(
			case (g_LBTools_BH.f_getPivotIndex g_LBTools_BH.bFootL.controller) of
			(
				1: chkbtn_ankle.checked = true
				2: chkbtn_backLeft.checked = true
				3: chkbtn_backMiddle.checked = true
				4: chkbtn_backRight.checked = true
				5: chkbtn_frontLeft.checked = true
				6: chkbtn_frontMiddle.checked = true
				7: chkbtn_frontRight.checked = true
				default: f_uncheckAll()
			)
		)
		
		on leftFootPivots_Rollout close do
		(
			g_LBTools_BH.SelectDisplay_Rollout.chkbtn_pivotsLFoot.checked = false
		)
		
		on chkbtn_ankle changed state do (f_updateButton state chkbtn_ankle 1)
		on chkbtn_backLeft changed state do (f_updateButton state chkbtn_backLeft 2)
		on chkbtn_backMiddle changed state do (f_updateButton state chkbtn_backMiddle 3)
		on chkbtn_backRight changed state do (f_updateButton state chkbtn_backRight 4)
		on chkbtn_frontLeft changed state do (f_updateButton state chkbtn_frontLeft 5)
		on chkbtn_frontMiddle changed state do (f_updateButton state chkbtn_frontMiddle 6)
		on chkbtn_frontRight changed state do (f_updateButton state chkbtn_frontRight 7)	
	) --// end rollout
	
	rollout rightFootPivots_Rollout "R-Foot"
	(
		local lrg = 16

		checkbutton chkbtn_frontLeft "" across:3 width:lrg height:lrg align:#left offset:[0,0] tooltip:"front left"
		checkbutton chkbtn_frontMiddle "" width:lrg height:lrg offset:[0,0] tooltip:"front middle"
		checkbutton chkbtn_frontRight "" width:lrg height:lrg offset:[3,0] tooltip:"front right"
		checkbutton chkbtn_ankle "" width:lrg height:lrg align:#left offset:[39,0] tooltip:"ankle"
		checkbutton chkbtn_backLeft "" across:3 width:lrg height:lrg align:#left offset:[0,0] tooltip:"back left"
		checkbutton chkbtn_backMiddle "" width:lrg height:lrg offset:[0,0] tooltip:"back middle"
		checkbutton chkbtn_backRight "" width:lrg height:lrg offset:[3,0] tooltip:"back right"

		fn f_uncheckAll =
		(
			chkbtn_ankle.checked = false
			chkbtn_backLeft.checked = false
			chkbtn_backMiddle.checked = false
			chkbtn_backRight.checked = false
			chkbtn_frontLeft.checked = false
			chkbtn_frontMiddle.checked = false
			chkbtn_frontRight.checked = false
		)
		
		fn f_updateButton state chkbtn index =
		(
			f_uncheckAll();
			if state then
			(
				if (g_LBTools_BH.f_changePivot g_LBTools_BH.bFootR.controller index) then
				(
					chkbtn.checked = true
				)
			)
			else
			(
				if (g_LBTools_BH.f_getPivotIndex g_LBTools_BH.bFootR.controller) > 0 then
				(
					chkbtn.checked = true
				)
			)
		)
	
		on rightFootPivots_Rollout open do
		(
			case (g_LBTools_BH.f_getPivotIndex g_LBTools_BH.bFootR.controller) of
			(
				1: chkbtn_ankle.checked = true
				2: chkbtn_backLeft.checked = true
				3: chkbtn_backMiddle.checked = true
				4: chkbtn_backRight.checked = true
				5: chkbtn_frontLeft.checked = true
				6: chkbtn_frontMiddle.checked = true
				7: chkbtn_frontRight.checked = true
				default: f_uncheckAll()
			)
		)
		
		on rightFootPivots_Rollout close do
		(
			g_LBTools_BH.SelectDisplay_Rollout.chkbtn_pivotsRFoot.checked = false
		)
		
		on chkbtn_ankle changed state do (f_updateButton state chkbtn_ankle 1)
		on chkbtn_backLeft changed state do (f_updateButton state chkbtn_backLeft 2)
		on chkbtn_backMiddle changed state do (f_updateButton state chkbtn_backMiddle 3)
		on chkbtn_backRight changed state do (f_updateButton state chkbtn_backRight 4)
		on chkbtn_frontLeft changed state do (f_updateButton state chkbtn_frontLeft 5)
		on chkbtn_frontMiddle changed state do (f_updateButton state chkbtn_frontMiddle 6)
		on chkbtn_frontRight changed state do (f_updateButton state chkbtn_frontRight 7)
	) --// end rollout
	
	rollout Keys_Rollout "Keys"
	(
		group "Add / Remove"
		(
			button btn_keyAllParts "Key All Parts" width:160 height:20 align:#center tooltip:"Add new keys to ALL Biped parts at the current slider time. Alternatively, use SHIFT or CONTROL to apply to across the time range."
			button btn_keySelectedParts "Key Selected Parts" width:160 height:20 align:#center tooltip:"Add new keys to SELECTED Biped parts at the current slider time. Alternatively, use SHIFT or CONTROL to apply to across the time range."
			button btn_deleteAllkeys "Delete All Keys" width:160 height:20 align:#center tooltip:"Remove all keys on ALL Biped parts."
			button btn_deleteFrameKeys "Delete Frame" width:78 height:20 align:#left offset:[3,0] across:2 tooltip:"Remove keys on SELECTED Biped parts at the current slider time. Alternatively, use SHIFT or CONTROL to remove keys on ALL Biped parts."
			button btn_deleteRangeKeys "Delete Range" width:78 height:20 align:#right offset:[-3,0] tooltip:"Remove keys on SELECTED Biped parts across the time range. Alternatively, use SHIFT or CONTROL to remove keys on ALL Biped parts."
			button btn_deleteBetweenKeys "Delete Between" width:115 height:20 align:#left offset:[3,0] across:2 tooltip:"Remove keys on SELECTED Biped parts across the time range between the spinner KEEP step value. Alternatively, use SHIFT or CONTROL to remove keys on ALL Biped parts."
			spinner spn_keepStep "" type:#integer range:[2,100,2] fieldwidth:30 align:#right offset:[-3,2] tooltip:"Step value for keeping keys every Nth frame"
		)
		group "TCB Continuity"
		(
			button btn_TCB0 "0" align:#center width:24 height:20 across:6 offset:[2,0] tooltip:"Set keys to 0 Continuity for SELECTED Biped parts at current slider time. Alternatively, hold SHIFT or CONTROL to apply across the time range."
			button btn_TCB5 "5" align:#center width:24 height:20 offset:[2,0] tooltip:"Set keys to 5 Continuity for SELECTED Biped parts at current slider time. Alternatively, hold SHIFT or CONTROL to apply across the time range."
			button btn_TCB10 "10" align:#center width:24 height:20 offset:[2,0] tooltip:"Set keys to 10 Continuity for SELECTED Biped parts at current slider time. Alternatively, hold SHIFT or CONTROL to apply across the time range."
			button btn_TCB15 "15" align:#center width:24 height:20 offset:[2,0] tooltip:"Set keys to 15 Continuity for SELECTED Biped parts at current slider time. Alternatively, hold SHIFT or CONTROL to apply across the time range."
			button btn_TCB20 "20" align:#center width:24 height:20 offset:[2,0] tooltip:"Set keys to 20 Continuity for SELECTED Biped parts at current slider time. Alternatively, hold SHIFT or CONTROL to apply across the time range."
			button btn_TCB25 "25" align:#center width:24 height:20 offset:[2,0] tooltip:"Set keys to 25 Continuity for SELECTED Biped parts at current slider time. Alternatively, hold SHIFT or CONTROL to apply across the time range."
		)
		
		fn f_addKeysToAllParts atTime:sliderTime =
		(
			try
			(
				biped.addNewKey g_LBTools_BH.bRoot.controller.vertical atTime
				biped.addNewKey g_LBTools_BH.bRoot.controller.horizontal atTime
				biped.addNewKey g_LBTools_BH.bRoot.controller.turning atTime
				biped.addNewKey g_LBTools_BH.bPelvis.controller atTime
				biped.addNewKey g_LBTools_BH.bSpine.controller atTime
				biped.addNewKey g_LBTools_BH.bHead.controller atTime
				biped.addNewKey g_LBTools_BH.bHandL.controller atTime
				biped.addNewKey g_LBTools_BH.bHandR.controller atTime
				biped.addNewKey g_LBTools_BH.bFootL.controller atTime
				biped.addNewKey g_LBTools_BH.bFootR.controller atTime
				biped.addNewKey g_LBTools_BH.bTail.controller atTime
			)catch()
		)

		fn f_addKeysToSelectedParts part atTime:sliderTime =
		(
			if part == g_LBTools_BH.bRoot then
			(
				biped.addNewKey part.controller.vertical atTime
				biped.addNewKey part.controller.horizontal atTime
				biped.addNewKey part.controller.turning atTime
			)
			else
			(
				biped.addNewKey part.controller atTime
			)
		)
		
		fn f_deleteAllFrameKeys atTime:sliderTime =
		(
			try(deleteKey g_LBTools_BH.bRoot.controller.vertical.controller (getKeyIndex g_LBTools_BH.bRoot.controller.vertical.controller atTime))catch()
			try(deleteKey g_LBTools_BH.bRoot.controller.horizontal.controller (getKeyIndex g_LBTools_BH.bRoot.controller.horizontal.controller atTime))catch()
			try(deleteKey g_LBTools_BH.bRoot.controller.turning.controller (getKeyIndex g_LBTools_BH.bRoot.controller.turning.controller atTime))catch()
			try(deleteKey g_LBTools_BH.bPelvis.controller (getKeyIndex g_LBTools_BH.bPelvis.controller atTime))catch()
			try(deleteKey g_LBTools_BH.bSpine.controller (getKeyIndex g_LBTools_BH.bSpine.controller atTime))catch()
			try(deleteKey g_LBTools_BH.bHead.controller (getKeyIndex g_LBTools_BH.bHead.controller atTime))catch()
			try(deleteKey g_LBTools_BH.bHandL.controller (getKeyIndex g_LBTools_BH.bHandL.controller atTime))catch()
			try(deleteKey g_LBTools_BH.bHandR.controller (getKeyIndex g_LBTools_BH.bHandR.controller atTime))catch()
			try(deleteKey g_LBTools_BH.bFootL.controller (getKeyIndex g_LBTools_BH.bFootL.controller atTime))catch()
			try(deleteKey g_LBTools_BH.bFootR.controller (getKeyIndex g_LBTools_BH.bFootR.controller atTime))catch()
			try(deleteKey g_LBTools_BH.bTail.controller (getKeyIndex g_LBTools_BH.bTail.controller atTime))catch()
		)
		
		fn f_deleteBipedPartKeys part atTime:sliderTime =
		(
			deselectKeys part.controller
			selectKeys part.controller atTime
			if part == g_LBTools_BH.bRoot then
			(
				deleteKeys part.controller.vertical.controller.keys #selection
				deleteKeys part.controller.horizontal.controller.keys #selection
				deleteKeys part.controller.turning.controller.keys #selection
			)
			else
			(
				deleteKeys part.controller.keys #selection
			)
		)
		
		on btn_keyAllParts pressed do
		(
			try
			(
				if g_LBTools_BH.bRoot.controller.figureMode == false then
				(
					undo "KeyAllParts" on
					(
						if keyboard.shiftPressed OR keyboard.controlPressed then
						(
							for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
							(	
								f_addKeysToAllParts atTime:t
							)
						)
						else --// Default method, add keys at current slider time
						(
							f_addKeysToAllParts()
						)
					)
				)
				else
				(
					messageBox "Error: Exit figure mode first." title:"Biped Helper"
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_keySelectedParts pressed do
		(
			try
			(
				if g_LBTools_BH.bRoot.controller.figureMode == false then
				(
					undo "KeySelectedParts" on
					(
						if selection.count > 0 then
						(
							local a_bipedParts = selection as array
							if keyboard.shiftPressed OR keyboard.controlPressed then
							(
								for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
								(
									for b in a_bipedParts do
									(
										f_addKeysToSelectedParts b atTime:t
									)
								)
							)
							else --// Default method, add keys at current slider time
							(
								for b in a_bipedParts do
								(
									f_addKeysToSelectedParts b
								)
							)
						)
						else
						(
							messageBox "Error: Select one or more Biped parts first." title:"Biped Helper"
						)
					)
				)
				else
				(
					messageBox "Error: Exit figure mode first." title:"Biped Helper"
				)
			)catch(g_LBTools_BH.f_initBipedParts())	
		)
		
		on btn_deleteAllkeys pressed do
		(
			try
			(
				if g_LBTools_BH.bRoot.controller.figureMode == false then
				(
					if (queryBox "Continue to delete all Biped animation?" title:"Biped Helper") do
					(
						biped.clearAllAnimation g_LBTools_BH.bRoot.controller
						g_LBTools_BH.bRoot.controller.figureMode = true
						g_LBTools_BH.bRoot.controller.figureMode = false
					)
				)
				else
				(
					messageBox "Error: Exit figure mode first." title:"Biped Helper"
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
			
		on btn_deleteFrameKeys pressed do
		(
			try
			(
				if g_LBTools_BH.bRoot.controller.figureMode == false then
				(
					undo "DeleteFrameKeys" on
					(
						if keyboard.shiftPressed OR keyboard.controlPressed then
						(
							f_deleteAllFrameKeys()
						)
						else --// Default method, removes keys only on selected parts
						(
							if selection.count > 0 then
							(
								local a_bipedParts = selection as array
								for b in a_bipedParts do
								(
									f_deleteBipedPartKeys b
								)
							)
							else
							(
								messageBox "Error: Select one or more Biped parts first." title:"Biped Helper"
							)
						)
					)
				)
				else
				(
					messageBox "Error: Exit figure mode first." title:"Biped Helper"
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_deleteRangeKeys pressed do
		(
			try
			(
				if g_LBTools_BH.bRoot.controller.figureMode == false then
				(
					undo "DeleteRangeKeys" on
					(
						if keyboard.shiftPressed OR keyboard.controlPressed then
						(
							for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
							(
								f_deleteAllFrameKeys atTime:t
							)
						)
						else --// Default method, removes keys only on selected parts
						(
							if selection.count > 0 then
							(
								local a_bipedParts = selection as array
								for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
								(
									for b in a_bipedParts do
									(
										f_deleteBipedPartKeys b atTime:t
									)
								)
							)
							else
							(
								messageBox "Error: Select one or more Biped parts first." title:"Biped Helper"
							)
						)
					)
				)
				else
				(
					messageBox "Error: Exit figure mode first." title:"Biped Helper"
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_deleteBetweenKeys pressed do
		(
			try
			(
				if g_LBTools_BH.bRoot.controller.figureMode == false then
				(
					undo "DeleteBetweenKeys" on
					(
						--// Delete Between will ignore the time range step size value, so it uses the keep step value instead.
						if keyboard.shiftPressed OR keyboard.controlPressed then
						(
							for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend while not keyboard.escPressed do
							(
								if (mod t.frame spn_keepStep.value) != 0 then
								(
									f_deleteAllFrameKeys atTime:t
								)
							)
						)
						else --// Default method, removes keys only on selected parts
						(
							if selection.count > 0 then
							(
								local a_bipedParts = selection as array
								for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend while not keyboard.escPressed do
								(
									if (mod t.frame spn_keepStep.value) != 0 then
									(
										for b in a_bipedParts do
										(
											f_deleteBipedPartKeys b atTime:t
										)
									)
								)
							)
							else
							(
								messageBox "Error: Select one or more Biped parts first." title:"Biped Helper"
							)
						)
					)
				)
				else
				(
					messageBox "Error: Exit figure mode first." title:"Biped Helper"
				)
			)catch(g_LBTools_BH.f_initBipedParts())	
		)
		
		on btn_TCB0 pressed do
		(
			g_LBTools_BH.f_evaluateTCBprocess 0.0
		)
		
		on btn_TCB5 pressed do
		(
			g_LBTools_BH.f_evaluateTCBprocess 5.0
		)
		
		on btn_TCB10 pressed do
		(
			g_LBTools_BH.f_evaluateTCBprocess 10.0
		)
		
		on btn_TCB15 pressed do
		(
			g_LBTools_BH.f_evaluateTCBprocess 15.0
		)
		
		on btn_TCB20 pressed do
		(
			g_LBTools_BH.f_evaluateTCBprocess 20.0
		)
		
		on btn_TCB25 pressed do
		(
			g_LBTools_BH.f_evaluateTCBprocess 25.0
		)
	) --// end rollout

	rollout LimbSpaces_Rollout "Limb Spaces"
	(
		group "Arm Spaces"
		(
			button btn_bothArmsObject "Both Arms Object" align:#center width:160 height:20 tooltip:"Set sliding keys (OBJECT space) for both arms. Hold SHIFT to set a planted key."
			button btn_LarmObject "L-Arm Object" align:#left across:2 width:78 height:20 offset:[3,0] tooltip:"Set a sliding key (OBJECT space) for the left arm. Hold SHIFT to set a planted key."
			button btn_RarmObject "R-Arm Object" align:#right width:78 height:20 offset:[-3,0] tooltip:"Set a sliding key (OBJECT space) for the right arm. Hold SHIFT to set a planted key."
			button btn_bothArmsBody "Both Arms Body" align:#center width:160 height:20 tooltip:"Set free keys (BODY space) for both arms."
			button btn_LarmBody "L-Arm Body" align:#left across:2 width:78 height:20 offset:[3,0] tooltip:"Set a free key (BODY space) for the left arm."
			button btn_RarmBody "R-Arm Body" align:#right width:78 height:20 offset:[-3,0] tooltip:"Set a free key BODY space) for the right arm."
		)
		group "Leg Spaces"
		(
			button btn_bothLegsObject "Both Legs Object" align:#center width:160 height:20 tooltip:"Set sliding keys (OBJECT space) for both legs. Hold SHIFT to set a planted key."
			button btn_LlegObject "L-Leg Object" align:#center across:2 width:78 height:20 tooltip:"Set a sliding key (OBJECT space) for the left leg. Hold SHIFT to set a planted key."
			button btn_RlegObject "R-Leg Object" align:#center width:78 height:20 tooltip:"Set a sliding key (OBJECT space) for the right leg. Hold SHIFT to set a planted key."
			button btn_bothLegsBody "Both Legs Body" align:#center width:160 height:20 tooltip:"Set free keys (BODY space) for both legs."
			button btn_LlegBody "L-Leg Body" align:#center across:2 width:78 height:20 tooltip:"Set a free key (BODY space) for the left leg."
			button btn_RlegBody "R-Leg Body" align:#center width:78 height:20 tooltip:"Set a free key (BODY space) for the right leg."
		)
		group "Store and Toggle Spaces"
		(
			button btn_limbObject "Limb Object" across:2 align:#left width:78 height:20 offset:[3,0] tooltip:"Set all keys for any selected limbs to OBJECT space."
			button btn_storeSpaces "Store Spaces" align:#right width:78 height:20 offset:[-3,0] tooltip:"Temporarily store all limb BODY/OBJECT space settings then set all limb keys to use BODY space."
			button btn_limbBody "Limb Body" across:2 align:#left width:78 height:20 offset:[3,0] tooltip:"Set all keys for any selected limbs to BODY space."
			button btn_loadSpaces "Load Spaces" align:#right width:78 height:20 offset:[-3,0] tooltip:"Load all limb BODY/OBJECT space settings based on a previous Store Spaces operation."
		)
		
		on btn_bothArmsObject pressed do
		(
			try(
				if keyboard.shiftPressed then
				(
					biped.setPlantedKey g_LBTools_BH.bHandL
					biped.setPlantedKey g_LBTools_BH.bHandR					
				)
				else
				(
					biped.setSlidingKey g_LBTools_BH.bHandL
					biped.setSlidingKey g_LBTools_BH.bHandR
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
			
		on btn_LarmObject pressed do
		(
			try(
				if keyboard.shiftPressed then
				(
					biped.setPlantedKey g_LBTools_BH.bHandL
				)
				else
				(
					biped.setSlidingKey g_LBTools_BH.bHandL
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_LarmBody pressed do
		(
			try
			(
				biped.setFreeKey g_LBTools_BH.bHandL
			)catch(g_LBTools_BH.f_initBipedParts())
		)
	
		on btn_bothArmsBody pressed do
		(
			try
			(
				biped.setFreeKey g_LBTools_BH.bHandL
			)catch(g_LBTools_BH.f_initBipedParts())
			try
			(
				biped.setFreeKey g_LBTools_BH.bHandR
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_RarmObject pressed do
		(
			try(
				if keyboard.shiftPressed then
				(
					biped.setPlantedKey g_LBTools_BH.bHandR
				)
				else
				(
					biped.setSlidingKey g_LBTools_BH.bHandR
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_RarmBody pressed do
		(
			try
			(
				biped.setFreeKey g_LBTools_BH.bHandR
			)catch(g_LBTools_BH.f_initBipedParts())
		)	
		
		on btn_bothLegsObject pressed do
		(
			try(
				if keyboard.shiftPressed then
				(
					biped.setPlantedKey g_LBTools_BH.bFootL
					biped.setPlantedKey g_LBTools_BH.bFootR
				)
				else
				(
					biped.setSlidingKey g_LBTools_BH.bFootL
					biped.setSlidingKey g_LBTools_BH.bFootR
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
			
		on btn_LlegObject pressed do
		(
			try(
				if keyboard.shiftPressed then
				(
					biped.setPlantedKey g_LBTools_BH.bFootL
				)
				else
				(
					biped.setSlidingKey g_LBTools_BH.bFootL
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_LlegBody pressed do
		(
			try
			(
				biped.setFreeKey g_LBTools_BH.bFootL
			)catch(g_LBTools_BH.f_initBipedParts())
		)
	
		on btn_bothLegsBody pressed do
		(
			try(
				biped.setFreeKey g_LBTools_BH.bFootL
				biped.setFreeKey g_LBTools_BH.bFootR
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_RlegObject pressed do
		(
			try(
				if keyboard.shiftPressed then
				(
					biped.setPlantedKey g_LBTools_BH.bFootR
				)
				else
				(
					biped.setSlidingKey g_LBTools_BH.bFootR
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_RlegBody pressed do
		(
			try
			(
				biped.setFreeKey g_LBTools_BH.bFootR
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_storeSpaces pressed do
		(
			try
			(
				g_LBTools_BH.a_storedSpaces = g_LBTools_BH.f_storeSpaces()
			)catch(g_LBTools_BH.f_initBipedParts())
		)

		on btn_loadSpaces pressed do
		(
			try
			(
				undo "LoadSpaces" on
				(
					g_LBTools_BH.f_loadSpaces g_LBTools_BH.a_storedSpaces
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)	

		on btn_limbBody pressed do
		(
			try(
				undo "LimbBody" on
				(
					g_LBTools_BH.f_limbBody()
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_limbObject pressed do
		(
			try(
				undo "LimbObject" on
				(
					g_LBTools_BH.f_limbObject()
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
	) --// end rollout
	
	rollout BipFiles_Rollout "Load / Save Animation"
	(
		group ""
		(
			button btn_loadAnimDialog "Load Dialog" align:#left width:78 height:20 offset:[3,0] across:2 tooltip:"Load an existing BIP file as the current animation via a dialog for the selected biped."
			button btn_saveAnimDialog "Save Dialog" align:#right width:78 height:20 offset:[-3,0] tooltip:"Save the current animation as a BIP file via a dialog for the selected biped."
			button btn_loadAnimQuick "Load Quick" align:#left width:78 height:20 offset:[3,0] across:2 tooltip:"Load a previously saved animation onto the selected biped."
			button btn_saveAnimQuick "Save Quick" align:#right width:78 height:20 offset:[-3,0] tooltip:"Save the current animation for the selected biped."
		)
			
		on btn_loadAnimDialog pressed do
		(
			try
			(
				biped.loadBipFileDlg g_LBTools_BH.bRoot.controller
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_saveAnimDialog pressed do
		(
			try
			(
				biped.saveBipFileDlg g_LBTools_BH.bRoot.controller
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_loadAnimQuick pressed do
		(
			try
			(
				if (doesFileExist g_LBTools_BH.quickBipFile) then
				(
					biped.loadBipFile g_LBTools_BH.bRoot.controller g_LBTools_BH.quickBipFile
					messagebox "Animation loaded successfully.\n" title:"Biped Helper"
				)
				else
				(
					messagebox "Error: No previously saved animation found." title:"Biped Helper"
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_saveAnimQuick pressed do
		(
			try
			(
				biped.saveBipFile g_LBTools_BH.bRoot.controller g_LBTools_BH.quickBipFile
				messagebox "Animation saved.\n" title:"Biped Helper"
			)catch(g_LBTools_BH.f_initBipedParts())
		)
	) --// end rollout
		
	rollout LayersRetargeting_Rollout "Biped Layers"
	(
		local ctrl = undefined
		
		fn f_filter obj = NOT (matchPattern obj.name pattern:(g_LBTools_BH.bRoot.name + "*"))
		
		group "Retargeting"
		(
			radiobuttons rdobtns_base labels:#("Biped's Base Layer") align:#left offsets:#([15,0]) default:1
			radiobuttons rdobtns_ref labels:#("Reference Biped") align:#left offsets:#([15,0]) default:0 enabled:false
			pickbutton pkbtn_refBiped ">  Pick a reference biped  <" width:173 height:20 align:#center offset:[-2,0] filter:f_filter message:"Pick another Biped to reference from..."
			checkbutton chkbtn_leftArm "L-Arm" across:4 width:40 height:20 align:#left
			checkbutton chkbtn_rightArm "R-Arm" width:40 height:20 align:#left
			checkbutton chkbtn_leftLeg "L-Leg" width:40 height:20 align:#right
			checkbutton chkbtn_rightLeg "R-Leg" width:40 height:20 align:#right
			button btn_update "Update" across:2 align:#left width:100 height:25
			checkbox chkbx_IKonly "IK Only" align:#right offset:[-5,5]
		)
		
		on LayersRetargeting_Rollout open do
		(
			ctrl = g_LBTools_BH.bRoot.controller
			
			local obj = biped.getRetargetRefBip ctrl
			if isValidNode obj then
			(
				pkbtn_refBiped.text = obj.name
				rdobtns_ref.enabled = true
			)
			else
			(
				pkbtn_refBiped.object = undefined
			)

			if g_LBTools_BH.retargetingOption == 1 then
			(
				rdobtns_base.state = 1
				rdobtns_ref.state = 0
			)
			else if g_LBTools_BH.retargetingOption == 2 then
			(
				rdobtns_base.state = 0
				rdobtns_ref.state = 1
			)
				
			chkbtn_leftArm.checked = biped.getLimbRetargetState g_LBTools_BH.bHandL
			chkbtn_rightArm.checked = biped.getLimbRetargetState g_LBTools_BH.bHandR
			chkbtn_leftLeg.checked = biped.getLimbRetargetState g_LBTools_BH.bFootL
			chkbtn_rightLeg.checked = biped.getLimbRetargetState g_LBTools_BH.bFootR
		)
		
		on LayersRetargeting_Rollout close do
		(
			g_LBTools_BH.Layers_Rollout.chkbtn_retargeting.state = false
		)
		
		on pkbtn_refBiped picked obj do
		(
			if isValidNode obj AND classof obj == Biped_Object then
			(
				local rootObj = g_LBTools_BH.f_findHierarchyRoot obj
				if rootObj != g_LBTools_BH.bRoot then
				(
					pkbtn_refBiped.text = rootObj.name
					rdobtns_base.state = 0
					rdobtns_ref.state = 1
					rdobtns_ref.enabled = true
					biped.setRetargetRefBip ctrl rootObj
				)
				else
				(
					messagebox "Error: Cannot be the same Biped." title:"Biped Layers"
				)
			)
			else
			(
				messagebox "Error: Please pick a Biped part." title:"Biped Layers"
			)
		)
		
		on rdobtns_base changed state do
		(
			if state == 1 then
			(
				rdobtns_ref.state = 0
				g_LBTools_BH.retargetingOption = 1
			)
		)
		
		on rdobtns_ref changed state do
		(
			if state == 1 then
			(
				rdobtns_base.state = 0
				g_LBTools_BH.retargetingOption = 2
			)
		)
		
		on chkbtn_leftArm changed state do
		(
			try
			(
				biped.setLimbRetargetState g_LBTools_BH.bHandL state
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on chkbtn_rightArm changed state do
		(
			try
			(
				biped.setLimbRetargetState g_LBTools_BH.bHandR state
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on chkbtn_leftLeg changed state do
		(
			try
			(
				biped.setLimbRetargetState g_LBTools_BH.bFootL state
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on chkbtn_rightLeg changed state do
		(
			try
			(
				biped.setLimbRetargetState g_LBTools_BH.bFootR state
			)catch(g_LBTools_BH.f_initBipedParts())
		)
			
		on btn_update pressed do
		(
			try
			(
				if rdobtns_base.state == 1 then
				(
					biped.retargetToBaseLayer ctrl chkbx_IKonly.state
				)
				else if rdobtns_ref.state == 1 then
				(
					local obj = biped.getRetargetRefBip ctrl
					if isValidNode obj then
					(
						biped.retargetToReferenceBiped ctrl chkbx_IKonly.state 
					)
				)
				redrawViews()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
	) --// end rollout
	
	rollout Layers_Rollout "Layers"
	(
		local a_layerNames = #()
		local a_layersActiveState = #()
		local ctrl = undefined
		local numLayers = 0
		local layerIndex = 1
		local retargetingUIoffset = [220,540]

		group ""
		(
			checkbox chkbx_active "Active" align:#left tooltip:"Toggle selected layer on or off."
			checkbutton chkbtn_solo "Solo" width:50 height:20 align:#left across:3 offset:[3,0] tooltip:"Activate only the selected layer."
			button btn_allOn "All On" width:50 height:20 align:#center offset:[0,0] tooltip:"Toggle all layers on."
			button btn_allOff "All Off" width:50 height:20 align:#right offset:[-3,0] tooltip:"Toggle all layers off."
			button btn_moveDown "Move Down" width:78 height:20 align:#left across:2 offset:[3,0] tooltip:"Move one layer down."
			button btn_moveUp "Move Up" width:78 height:20 align:#right offset:[-3,0] tooltip:"Move one layer up."
			edittext edtxt_index "" readonly:true across:3 align:#left width:25 offset:[0,2]
			dropDownList ddl_layers "" width:110 align:#center offset:[3,0]
			button btn_refreshList "R" width:20 height:20 align:#right offset:[-2,0] tooltip:"Refresh the list by syncing with the Biped panel."
			edittext edtxt_LayerName "" fieldwidth:103 align:#left across:2 offset:[0,2]
			button btn_renameLayer "Rename" width:50 height:20 align:#right offset:[-3,0]
			button btn_addLayer "Add" width:50 height:20 align:#left across:3 offset:[3,0]
			button btn_deleteLayer "Delete" width:50 height:20 align:#center
			button btn_collapseLayers "Collapse" width:50 height:20 align:#right offset:[-3,0]
			button btn_snapSetKey "Snap & Set Key" align:#center width:160 height:20 tooltip:"Sets a snap key for the current layer to match the base layer."
			checkbutton chkbtn_retargeting "Retargeting UI  >>" align:#center width:160 height:20 tooltip:"Opens a separate dialog for controlling IK retargeting."
		)
			
		fn f_getDisplayNames =
		(
			local a_displayNames = deepCopy a_layerNames
			for i = 1 to a_displayNames.count-1 do
			(
				if (biped.getLayerActive ctrl i) then
				(
					a_displayNames[i+1] = a_layerNames[i+1]
				)
				else
				(
					a_displayNames[i+1] = "* " + a_layerNames[i+1]
				)	
			)
			
			a_displayNames
		)
		
		fn f_syncBipedPanel =
		(
			numLayers = biped.numLayers ctrl
			layerIndex = biped.getCurrentLayer ctrl
		)
		
		fn f_updateLayers =
		(
			if chkbtn_solo.state then
			(
				btn_addLayer.enabled = false
				btn_deleteLayer.enabled = false
				btn_collapseLayers.enabled = false
				btn_renameLayer.enabled = false
				btn_allOn.enabled = false
				btn_allOff.enabled = false
				btn_moveDown.enabled = false
				btn_moveUp.enabled = false
				ddl_layers.enabled = false
				btn_refreshList.enabled = false
			)
			else
			(
				btn_addLayer.enabled = true
				btn_deleteLayer.enabled = true
				btn_collapseLayers.enabled = true
				btn_renameLayer.enabled = true
				btn_allOn.enabled = true
				btn_allOff.enabled = true
				btn_moveUp.enabled = true
				btn_moveDown.enabled = true
				ddl_layers.enabled = true
				btn_refreshList.enabled = true
				chkbx_active.enabled = true
				chkbx_active.checked = true
				chkbtn_solo.enabled = true
			)
			
			if g_LBTools_BH.f_isValidBiped() then
			(
				ctrl = g_LBTools_BH.bRoot.controller
				a_layerNames = #(("# "+(biped.getLayerName ctrl 0)+" #"))
				numLayers = biped.numLayers ctrl
				if numLayers > 0 then
				(
					layerIndex = biped.getCurrentLayer ctrl
					edtxt_index.text = layerIndex as string
					for i = 1 to numLayers do
					(
						append a_layerNames (biped.getLayerName ctrl i)
					)
				)
				else
				(
					layerIndex = 0
					edtxt_index.text = "0"
					chkbx_active.enabled = false
				)

				if layerIndex > 0 then
				(
					chkbx_active.checked = (biped.getLayerActive ctrl layerIndex)
				)
				else
				(
					chkbx_active.enabled = false
				)
				
				ddl_layers.items = f_getDisplayNames()
				ddl_layers.selection = layerIndex + 1
			)
			else
			(
				g_LBTools_BH.f_initBipedParts()
			)
			
			if layerIndex == numLayers then
			(
				btn_moveUp.enabled = false
			)
			if layerIndex == 0 then
			(
				btn_moveDown.enabled = false
			)
			if numLayers == 0 then
			(
				btn_deleteLayer.enabled = false
				btn_collapseLayers.enabled = false
			)
			else
			(
				if NOT chkbtn_solo.state then
				(
					btn_deleteLayer.enabled = true
					btn_collapseLayers.enabled = true
				)
			)
			redrawViews()
		)
		
		on Layers_Rollout open do
		(
			f_updateLayers()
		)
		
		on btn_renameLayer pressed do
		(
			try
			(
				undo "RenameLayer" on
				(
					local a_filtered = filterstring edtxt_layerName.text " "
					if a_filtered.count > 0 then
					(
						local index = biped.getCurrentLayer ctrl
						biped.setLayerName ctrl index edtxt_layerName.text
					)
					else
					(
						messagebox "Error: Must enter a valid layer name first." title:"Biped Helper"
						edtxt_layerName.text = ""
					)
				)
				f_updateLayers()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_addLayer pressed do
		(
			try
			(
				undo "AddLayer" on
				(
					local index = (biped.getCurrentLayer ctrl) + 1
					biped.createLayer ctrl index ("Layer " + index as string)
				)
				f_updateLayers()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_deleteLayer pressed do
		(
			try
			(
				if layerIndex != 0 then
				(
					undo "DeleteLayer" on
					(
						try(biped.deleteLayer ctrl layerIndex)catch(f_updateLayers())
					)
				)
				f_updateLayers()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_collapseLayers pressed do
		(
			try
			(
				local index = (biped.getCurrentLayer ctrl) - 1
				if index != -1 then
				(
					undo "CollapseLayers" on
					(
						biped.collapseAtLayer ctrl index
					)
				)
				f_updateLayers()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_snapSetKey pressed do
		(
			try
			(
				f_syncBipedPanel()
				if $selection.count == 1 then
				(
					biped.setSnapKey selection[1]
				)
				else
				(
					messagebox "Error: Select a single Biped node to perform a Snap & Set Key." title:"Biped Layers"
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on chkbtn_retargeting changed state do
		(
			try
			(
				if state then
				(
					try(destroyDialog g_LBTools_BH.LayersRetargeting_Rollout)catch()
					createDialog g_LBTools_BH.LayersRetargeting_Rollout width:200 pos:(g_LBTools_BH.BH_pos + g_LBTools_BH.layersRetargetingUIoffset) style:#(#style_toolwindow, #style_sysmenu)
				)
				else
				(
					try(destroyDialog g_LBTools_BH.LayersRetargeting_Rollout)catch()
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_moveDown pressed do
		(
			try
			(
				if g_LBTools_BH.f_isValidBiped() then
				(
					f_syncBipedPanel()
					layerIndex = layerIndex - 1
					if layerIndex >= 0 then
					(
						biped.setCurrentLayer ctrl layerIndex
						btn_moveUp.enabled = true
						if layerIndex == 0 then
						(
							btn_moveDown.enabled = false
						)
					)
					else
					(
						layerIndex = 0
					)
				)	
				f_updateLayers()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_moveUp pressed do
		(
			try
			(
				if g_LBTools_BH.f_isValidBiped() then
				(
					f_syncBipedPanel()
					layerIndex = layerIndex + 1
					if layerIndex <= numLayers then
					(
						biped.setCurrentLayer ctrl layerIndex
						btn_moveDown.enabled = true
						if layerIndex == numLayers then
						(
							btn_moveUp.enabled = false
						)
					)
					else
					(
						layerIndex = numLayers
						btn_moveUp.enabled = false
					)
				)
				f_updateLayers()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on chkbtn_solo changed state do
		(
			try
			(
				f_syncBipedPanel()
				if state then
				(					
					for i = 1 to numLayers do
					(
						local activeState = biped.getLayerActive ctrl i
						if activeState then
						(
							append a_layersActiveState true
						)
						else
						(
							append a_layersActiveState false
						)
					)
					
					for i = 1 to numLayers do
					(
						biped.setLayerActive ctrl i false
					)
					biped.setLayerActive ctrl layerIndex true
				)
				else
				(
					for i = 1 to numLayers do
					(
						biped.setLayerActive ctrl i a_layersActiveState[i]
					)
					a_layersActiveState = #()
				)
				f_updateLayers()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_allOn pressed do
		(
			try
			(
				f_syncBipedPanel()
				for i = 1 to numLayers do
				(
					biped.setLayerActive ctrl i true
				)
				f_updateLayers()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_allOff pressed do
		(
			try
			(
				f_syncBipedPanel()
				for i = 1 to numLayers do
				(
					biped.setLayerActive ctrl i false
				)
				f_updateLayers()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on chkbx_active changed state do
		(
			try
			(
				f_syncBipedPanel()
				if state then
				(
					biped.setLayerActive ctrl layerIndex true
				)
				else
				(
					biped.setLayerActive ctrl layerIndex false
				)
				f_updateLayers()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on ddl_layers selected val do
		(
			try
			(
				f_syncBipedPanel()
				local index = val - 1
				if index <= numLayers then
				(
					biped.setCurrentLayer ctrl index
				)
				f_updateLayers()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_refreshList pressed do
		(
			try
			(
				f_updateLayers()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
	) --// end rollout
	
	rollout PosePosture_Rollout "Pose / Posture"
	(
		local postureObjsFile = getDir #plugcfg + "\LBTools_BipedHelper_PostureObjs.txt"
		local TCBmode = undefined
		
		group "Paste Options"
		(
			checkbox chkbx_pasteOpposite "Paste with opposite limbs" align:#left offset:[0,0] tooltip:"Paste the pose or posture onto the opposite side for each limb."
			label lbl_COM "COM:" align:#left offset:[0,3] across:5
			checkbutton chkbtn_horizontal "H" align:#left offset:[-5,0] checked:true tooltip:"Paste horizontal COM data from the copy."
			checkbutton chkbtn_vertical "V" align:#left offset:[-12,0] checked:true tooltip:"Paste vertical COM data from the copy."
			checkbutton chkbtn_rotation "R" align:#left offset:[-20,0] checked:true tooltip:"Paste rotation COM data from the copy."
			checkbox chkbx_velocity "Velocity" align:#right offset:[0,3] tooltip:"Determines the values of the active COM tracks based on the previous trajectory of the COM through the scene. If there is no previous trajectory from which to gather the velocity data, the COM result is considered as absolute."
			radiobuttons rdobtns_tcb "Auto-Key TCB / IK Values:" labels:#("Default","Copied","Interp") columns:3 align:#left tooltip:#("Sets the TCB Ease To and Ease From to 0, and Tension, Continuity and Bias to 25.","Sets the TCB/IK values to match those from the copied data.","Sets the TCB values to an interpolation from the animation on which you are pasting.")
		)
		group "Pose (whole body)"
		(
			button btn_copyPose "Copy Pose" width:85 height:20 align:#left offset:[-5,0] across:2 tooltip:"Copy a single pose to a file for later pasting using the Paste Pose button."
			button btn_pastePose "Paste Pose" width:85 height:20 align:#right offset:[5,0] tooltip:"Paste a previously copied pose at the current slider time. Alternatively, use SHIFT or CONTROL to apply to across the time range."
		)
		group "Posture (selected parts only)"
		(
			button btn_copyPosture "Copy Posture" width:77 height:20 across:3 align:#left offset:[-3,0] tooltip:"Copy the Biped parts selection as a saved posture to a file for later pasting using the Paste Posture button."
			button btn_pastePosture "Paste Posture" width:77 height:20 align:#left offset:[22,0] tooltip:"Paste the previously saved posture onto the relevant biped parts at the current slider time. Alternatively, use SHIFT or CONTROL to apply to across the time range."
			button btn_selectPostureObjs "S" width:14 height:20 align:#right offset:[6,0] tooltip:"Select the Biped parts from which the posture was copied."
		)
		group "Mirror Limbs"
		(
			button btn_mirrorArmLeft "Mirror L-Arm" width:85 height:20 across:2 offset:[-5,0] align:#left tooltip:"Mirror the left arm posture to the right arm (including fingers)."
			button btn_mirrorArmRight "Mirror R-Arm" width:85 height:20 offset:[5,0] align:#right tooltip:"Mirror the right arm posture to the left arm (including fingers)."
			button btn_mirrorLegLeft "Mirror L-Leg" width:85 height:20 offset:[-5,0] across:2 align:#left tooltip:"Mirror the left leg posture to the right leg (including toes)."
			button btn_mirrorLegRight "Mirror R-Leg" width:85 height:20 offset:[5,0] align:#right tooltip:"Mirror the right leg posture to the left leg (including toes)."
		)
		group "Fingers Align"
		(
			button btn_straightenFingers "Straighten" width:85 height:20 align:#left offset:[-5,0] across:2 tooltip:"Straightens the last two finger controls for the selected fingers. If SHIFT or CONTROL is held all finger joints for the selected fingers are straightened."
			button btn_selectAllFingers "Select Fingers" width:85 height:20 align:#right offset:[5,0] tooltip:"Select all the finger controls on both hands."
		)
		group "Spine Align"
		(
			button btn_straightenSpine "Straighten" width:85 height:20 across:2 offset:[-5,0] align:#left tooltip:"Straightens the spine links to their default orientation."
			button btn_untwistSpine "Untwist" width:85 height:20 offset:[5,0] align:#right tooltip:"Untwists the spine links to align their orientation."
		)
		
		fn f_getTCBmode =
		(
			case rdobtns_tcb.state of
			(
				1: TCBmode = #pstdefault
				2: TCBmode = #pstcopied
				3: TCBmode = #pstinterp
			)
		)
		
		fn f_mirrorPosture a_bipObjs =
		(
			if isValidNode g_LBTools_BH.bRoot then
			(
				f_getTCBmode()
				local bRootCtrl = g_LBTools_BH.bRoot.controller
				biped.createCopyCollection bRootCtrl "MirrorTemp"
				local colIndex = biped.numCopyCollections bRootCtrl
				local copyCol = biped.getCopyCollection bRootCtrl colIndex
				biped.copyBipPosture bRootCtrl copyCol a_bipObjs #snapAuto
				local mirroreBH_pose = getCopy copyCol #posture 1
				with animate on
				(
					undo "MirrorPosture" on
					(
						biped.pasteBipPosture bRootCtrl mirroreBH_pose true TCBmode chkbtn_horizontal.checked chkbtn_vertical.checked chkbtn_rotation.checked chkbx_velocity.checked
					)
				)
				biped.deleteCopyCollection bRootCtrl colIndex
			)
		)
		
		fn f_storePostureObjs a_objs =
		(
			local txtFile = openFile postureObjsFile mode:"w"
			for obj in a_objs do
			(
				format "%\n" obj.name to:txtFile
			)
			close txtFile
		)
		
		fn f_retrievePostureObjs =
		(
			if doesFileExist postureObjsFile do
			(
				local a_objs = #()
				local txtFile = openFile postureObjsFile mode:"r"
				while NOT eof txtFile do
				(
					append a_objs (getNodeByName (readLine txtFile))
				)
				close txtFile
				a_objs
			)
		)
		
		on PosePosture_Rollout open do
		(
			if NOT (doesFileExist postureObjsFile) do
			(
				local txtFile = openFile postureObjsFile mode:"w"
				close txtFile
			)
		)
		
		on btn_copyPose pressed do
		(
			try
			(
				local bRootCtrl = g_LBTools_BH.bRoot.controller
				biped.createCopyCollection bRootCtrl "PoseGrabber"
				local colIndex = biped.numCopyCollections bRootCtrl
				local copyCol = biped.getCopyCollection bRootCtrl colIndex
				biped.copyBipPose bRootCtrl copyCol #snapAuto
				local copyBH_pose = getCopy copyCol #pose 1
				biped.saveCopyPasteFile bRootCtrl (getDir #plugcfg + "\LBTools_BipedHelper_PoseGrabbed.cpy")
				biped.deleteCopyCollection bRootCtrl colIndex
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_pastePose pressed do
		(
			try
			(
				local poseFile = getDir #plugcfg + "\LBTools_BipedHelper_PoseGrabbed.cpy"
				if doesFileExist poseFile then
				(
					f_getTCBmode()
					local bRootCtrl = g_LBTools_BH.bRoot.controller
					biped.loadCopyPasteFile bRootCtrl poseFile
					local colIndex = biped.numCopyCollections bRootCtrl
					local copyCol = biped.getCopyCollection bRootCtrl colIndex
					local copyBH_pose = getCopy copyCol #pose 1
					with animate on
					(
						undo "PastePose" on
						(
							if keyboard.shiftPressed OR keyboard.controlPressed then
							(
								for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
								(
									slidertime = t
									biped.pasteBipPose bRootCtrl copyBH_pose chkbx_pasteOpposite.checked TCBmode chkbtn_horizontal.checked chkbtn_vertical.checked chkbtn_rotation.checked chkbx_velocity.checked
								)
							)
							else
							(
								biped.pasteBipPose bRootCtrl copyBH_pose chkbx_pasteOpposite.checked TCBmode chkbtn_horizontal.checked chkbtn_vertical.checked chkbtn_rotation.checked chkbx_velocity.checked
							)
						)
					)
					local colIndex = biped.numCopyCollections bRootCtrl
					biped.deleteCopyCollection bRootCtrl colIndex
				)
				else
				(
					messageBox "Error: The pose file does not exist yet. Grab a pose first." title:"Biped Pose"
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_copyPosture pressed do
		(
			try
			(
				if selection.count > 0 then
				(
					local a_objs = selection as array
					--// Make sure only Biped parts are selected for copying by filtering out non-Biped parts
					a_objs = f_filterBipedParts a_objs
					if a_objs.count > 0 then
					(
						f_storePostureObjs a_objs
						if isValidNode g_LBTools_BH.bRoot then
						(
							local bRootCtrl = g_LBTools_BH.bRoot.controller
							biped.createCopyCollection bRootCtrl "CopyPosture"
							local colIndex = biped.numCopyCollections bRootCtrl
							local copyCol = biped.getCopyCollection bRootCtrl colIndex
							biped.copyBipPosture bRootCtrl copyCol a_objs #snapAuto
							biped.saveCopyPasteFile bRootCtrl (getDir #plugcfg + "\LBTools_BipedHelper_CopyPosture.cpy")
							biped.deleteCopyCollection bRootCtrl colIndex
						)
					)
					else
					(
						messageBox "Error: No Biped objects were selected." title:"Biped Helper"
					)
				)
				else
				(
					messageBox "Error: Select some Biped objects to copy." title:"Biped Helper"
				)
			)catch(g_LBTools_BH.f_initBipedParts())			
		)
		
		on btn_pastePosture pressed do
		(
			try
			(
				local postureFile = getDir #plugcfg + "\LBTools_BipedHelper_CopyPosture.cpy"
				if doesFileExist postureFile then
				(
					f_getTCBmode()
					local bRootCtrl = g_LBTools_BH.bRoot.controller
					biped.loadCopyPasteFile bRootCtrl postureFile
					local colIndex = biped.numCopyCollections bRootCtrl
					local copyCol = biped.getCopyCollection bRootCtrl colIndex
					local copyBH_posture = getCopy copyCol #posture 1
					if copyBH_posture != false then
					(
						with animate on
						(
							undo "PastePosture" on
							(
								if keyboard.shiftPressed OR keyboard.controlPressed then
								(
									for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
									(
										slidertime = t
										biped.pasteBipPosture bRootCtrl copyBH_posture chkbx_pasteOpposite.checked TCBmode chkbtn_horizontal.checked chkbtn_vertical.checked chkbtn_rotation.checked chkbx_velocity.checked
									)
								)
								else
								(
									biped.pasteBipPosture bRootCtrl copyBH_posture chkbx_pasteOpposite.checked TCBmode chkbtn_horizontal.checked chkbtn_vertical.checked chkbtn_rotation.checked chkbx_velocity.checked
								)
							)
						)
					)
					else
					(
						messagebox "Error: Ensure the Copy Collections for Posture doesn't contain a collection called \"CopyPosture\"." title:"Biped Helper"
					)
					local colIndex = biped.numCopyCollections bRootCtrl
					biped.deleteCopyCollection bRootCtrl colIndex
				)
				else
				(
					messageBox "Error: The pose file does not exist yet. Grab a pose first." title:"Biped Helper"
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_selectPostureObjs pressed do
		(
			try
			(
				local a_objs = f_retrievePostureObjs()
				if a_objs != undefined then
				(
					try
					(
						select a_objs
					)catch(messageBox "Error: Could not select the Biped parts used to copy the posture." title:"Biped Helper")
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_mirrorArmLeft pressed do
		(
			try
			(
				f_mirrorPosture (join g_LBTools_BH.a_armLeft g_LBTools_BH.a_fingersLeft)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_mirrorArmRight pressed do
		(
			try
			(
				f_mirrorPosture (join g_LBTools_BH.a_armRight g_LBTools_BH.a_fingersRight)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_mirrorLegLeft pressed do
		(
			try
			(
				f_mirrorPosture (join g_LBTools_BH.a_legLeft g_LBTools_BH.a_toesLeft)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_mirrorLegRight pressed do
		(
			try
			(
				f_mirrorPosture (join g_LBTools_BH.a_legRight g_LBTools_BH.a_toesRight)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_straightenFingers pressed do
		(
			try
			(
				with animate on
				(
					undo "StraightenFingers" on
					(
						if selection.count > 0 then
						(
							local a_selFingerParts = #()
							for obj in $selection where ((findItem g_LBTools_BH.a_fingersLeft obj) != 0) OR ((findItem g_LBTools_BH.a_fingersRight obj) != 0) do
							(
								append a_selFingerParts obj
							)
							local a_selFingers = #()
							for finger in a_selFingerParts do
							(
								local fingerRoot = g_LBTools_BH.f_findFingerRoot finger
								if keyboard.shiftPressed OR keyboard.controlPressed then
								(
									append a_selFingers fingerRoot
								)
								g_LBTools_BH.f_getFingerChildren fingerRoot a_nodes:a_selFingers
							)
							for obj in a_selFingers do
							(
								biped.setTransform g_LBTools_BH.bRoot #rotation obj.parent.transform.rotation false limb:obj
								biped.setSelectedKey obj.controller
							)
						)
						else
						(
							messageBox "Error: Select some Biped fingers to straighten." title:"Biped Helper"
						)
					)
				)
				redrawViews()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_selectAllFingers pressed do
		(
			try
			(
				clearSelection()
				select g_LBTools_BH.a_fingersLeft
				selectmore g_LBTools_BH.a_fingersRight
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_straightenSpine pressed do
		(
			try
			(
				with animate on
				(
					biped.zeroAll g_LBTools_BH.bSpine
				)
				redrawViews()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_untwistSpine pressed do
		(
			try
			(
				with animate on
				(
					biped.zeroTwist g_LBTools_BH.bSpine
				)
				redrawViews()
			)catch(g_LBTools_BH.f_initBipedParts())
		)
	) --// end rollout
	
	rollout Reference_Rollout "Reference"
	(
		group "Snapshot Skinned Meshes"
		(
			label lbl_sceneLayer "Scene layer to snapshot from:" align:#left
			edittext edtxt_sceneLayer "" align:#center width:162 offset:[-2,0] tooltip:"Enter the name of the scene layer containing skinned meshes."
			button btn_snapshot "Snap Current Frame" width:160 height:20 align:#center tooltip:"Creates an combined snapshot of all skinned meshes at the current slider time and makes it unobtrusive in the viewport."
			button btn_hideSnapshots "Hide/Unhide" align:#left width:78 height:20 offset:[3,0] across:2 tooltip:"Hides or unhides the SNAPSHOTS scene layer."
			button btn_deleteSnapshots "Delete" align:#right width:78 height:20 offset:[-3,0] tooltip:"Deletes the SNAPSHOTS scene layer and its contents."
		)		
		
		on Reference_Rollout open do
		(
			local snapshotLayerName = getIniSetting g_LBTools_BH.bhINIfile "SnapshotLayer" "LayerName"
			if snapshotLayerName == "" do
			(
				setIniSetting g_LBTools_BH.bhINIfile "SnapshotLayer" "LayerName" "*** BLANK ***"
				snapshotLayerName = "*** BLANK ***"
			)
			edtxt_sceneLayer.text = snapshotLayerName
		)
		
		on edtxt_sceneLayer changed val do
		(
			setIniSetting g_LBTools_BH.bhINIfile "SnapshotLayer" "LayerName" val
		)
		
		on btn_snapshot pressed do
		(
			try
			(
				local snapshotLayerName = getIniSetting g_LBTools_BH.bhINIfile "SnapshotLayer" "LayerName" 
				local layerMeshChar = layerManager.getLayerFromName snapshotLayerName
				-- // First make sure the scene layer exists
				if layerMeshChar != undefined then
				(
					--// Collect all nodes from the scene layer
					local a_nodes = #()
					layerMeshChar.nodes &a_nodes
					if a_nodes.count > 0 then
					(
						--// Collect only the skin meshes from the scene layer nodes array
						local a_skinNodes = #()
						for n in a_nodes do
						(
							if (g_LBTools_BH.f_findSkinModifier n) then
							(
								append a_skinNodes n
							)
						)
						if a_skinNodes.count > 0 then
						(
							with Animate off
							(
								--// Snapshot and attach all the skin mesh copies
								local snapMesh = snapShot a_skinNodes[1]
								for i = 2 to a_skinNodes.count do
								(
									attach snapMesh (snapshot a_skinNodes[i])
								)
								snapMesh.pivot = snapMesh.center
								-- // Create a new layer if this is the first snapshot
								if LayerManager.getLayerFromName "SNAPSHOTS" == undefined then
								(
									LayerManager.newLayerFromName "SNAPSHOTS"
								)
								-- // Add the new snapshot mesh to the snapshots layer, freeze it and set object properties transparency
								local layerSnapshots = LayerManager.getLayerFromName "SNAPSHOTS"
								layerSnapshots.addNode snapMesh
								freeze snapMesh
								snapMesh.backfacecull = on
								snapMesh.showFrozenInGray = off
								-- // Rename the snapshot mesh and add the slider time frame
								snapMesh.name = ("Snapshot :: " + sliderTime as string)
								-- // Assign a material which has a low opacity and grey colour
								snapMesh.material = meditMaterials[19]
								meditMaterials[19].Diffuse = color 150 150 150
								meditMaterials[19].ambient = color 150 150 150
								meditMaterials[19].opacity = 10
								meditMaterials[19].name = "See Through"
								snapMesh.wirecolor = color 150 150 150
							)
						)
						else
						(
							messageBox "Error: The snapshots scene layer should contain skinned meshes." title:"Biped Snapshot"
						)
					)
					else
					(
						messageBox "Error: No nodes were found the snapshots scene layer." title:"Biped Snapshot"
					)
				)
				else
				(
					messageBox "Error: Snapshots scene layer not found." title:"Biped Snapshot"
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_hideSnapshots pressed do
		(
			local layerSnapshots = layerManager.getLayerFromName "SNAPSHOTS"
			if layerSnapshots != undefined then
			(
				if layerSnapshots.isHidden then
				(
					layerSnapshots.isHidden = false
				)
				else
				(
					layerSnapshots.isHidden = true
				)
			)
		)		
		
		on btn_deleteSnapshots pressed do
		(
			local layerSnapshots = layerManager.getLayerFromName "SNAPSHOTS"
			if layerSnapshots != undefined then
			(
				layerSnapshots.nodes &nodesInLayer
				delete nodesInLayer
				layerManager.deleteLayerByName "SNAPSHOTS"
			)
		)
	) --// end rollout
		
	rollout AlignBake_Rollout "Align and Bake"
	(
		local bool_head = false, bool_root = true, bool_allLimbs = false, bool_both = false, bool_selLimbs = false, bool_nextSlideKey = false, bool_lastSlideKey = false
		local a_limbSel = #() --// Used by the selected limb toggles on the interface
		local propTargetObj = undefined
		
		group "Biped Alignment"
		(
			button btn_alignToOffset "Align To Offset" width:160 height:20 align:#center tooltip:"Aligns the root to start from the same position and orientation as a picked target object."
			button btn_alignToMatch "Align To Match" width:160 height:20 align:#center tooltip:"Aligns and re-animates all keys on the root to match a picked target object's motion."
		)
		group "Biped Baking"
		(
			radiobuttons rbtns_head "" labels:#("Head (to Look At Target)") align:#left default:0 tooltip:#("Bake keys onto the head based on an assigned Look At Target and removes the target object.")
			radiobuttons rbtns_rootLimbsBoth labels:#("Root", "All Limbs", "Both") align:#left default:1 offsets:#([0,0],[2,0],[2,0]) tooltip:#("Bake keys on the root bone across the full time range.", "Bake keys on all limbs based on Object bindings.", "Bake keys on both the root bone and all limbs.")
			radiobuttons rbtns_selLimbs labels:#("Selected Limbs:") columns:1 align:#left default:0 tooltip:"Allow specific limb selection for baking keys."
			checkbutton chkbtn_leftHand "L Hand" width:60 height:20 offset:[10,0] across:2 align:#center enabled:false tooltip:"Bake keys on the left hand between sliding keys."
			checkbutton chkbtn_rightHand "R Hand" width:60 height:20 offset:[-10,0] align:#center enabled:false tooltip:"Bake keys on the right hand between sliding keys."
			checkbutton chkbtn_leftFoot "L Foot" width:60 height:20 offset:[10,0] across:2 align:#center enabled:false tooltip:"Bake keys on the left foot between sliding keys."
			checkbutton chkbtn_rightFoot "R Foot" width:60 height:20 offset:[-10,0] align:#center enabled:false tooltip:"Bake keys on the right foot between sliding keys."
			label lbl_note1 "On a single selected limb starting" align:#left
			label lbl_note2 "from the slider time..." align:#left
			radiobuttons rbtns_nextSlideKey labels:#("Up to the NEXT sliding key") columns:1 align:#left default:0 tooltip:"Bake keys on a single selected limb from the first sliding key found after the slider time up to the NEXT sliding key (within range)."
			radiobuttons rbtns_lastSlideKey labels:#("Up to the LAST sliding key") columns:1 align:#left default:0 tooltip:"Bake keys on a single selected limb from the first sliding key found after the slider time up to the LAST sliding key (within range)."
			button btn_bakeAnimationKeys "Bake Animation Keys" width:160 height:20 align:#center tooltip:"Bake animation keys based on the mode selection above."
		)
		group "Prop Alignment"
		(
			label lbl_tgt "TGT:" align:#left across:2 offset:[0,3]
			pickbutton pkbtn_propTarget "Pick a TARGET" autoDisplay:true width:140 height:20 align:#right offset:[-3,0] tooltip:"Pick an object to use as the target for aligning or baking selected props."
			button btn_alignPropToTarget "Align To Target" width:110 height:20 align:#left offset:[3,0] across:2 tooltip:"Aligns the selected props to the picked target object and sets a key at the current slider time. Alternatively, use the Bake option to apply across the time range."
			checkbox chkbx_propBake "Bake" align:#right offset:[-5,3] tooltip:"Bake the alignment of the selected props across the chosen time range."
		)
		
		fn f_isProp obj =
		(
			if obj == g_LBTools_BH.bProp1 OR obj == g_LBTools_BH.bProp2 OR obj == g_LBTools_BH.bProp3 then
			(
				return true
			)
			false
		)
		
		on btn_alignToOffset pressed do
		(
			try
			(
				setCommandPanelTaskMode #create
				local bipObj = g_LBTools_BH.bRoot
				local pickedObj = pickObject message:"Pick a target object to align to..." rubberband:bipObj.center
				if pickedObj != undefined AND pickedObj != #escape then
				(
					select bipObj
					local pt1 = Point()
					local pt2 = Point()
					hide pt1
					hide pt2
					pt2.parent = pt1
					pt1.transform = bipObj.transform
					with animate on
					(
						for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
						(
							at time t (pt2.transform = bipObj.transform)
						)
					)
					pt1.transform = pickedObj.transform
					with animate on
					(
						local a_storedSpaces = g_LBTools_BH.f_storeSpaces()
						for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
						(
							sliderTime = t
							local verticalIndex = getKeyIndex bipObj[3][1].controller slidertime
							local horizontalIndex = getKeyIndex bipObj[3][2].controller slidertime
							local turningIndex = getKeyIndex bipObj[3][3].controller slidertime
							if verticalIndex > 0 OR horizontalIndex > 0 OR turningIndex > 0 then
							(
								at time t
								(
									bipObj.transform = pt2.transform
									if verticalIndex == 0 then
									(
										selectKeys bipObj[3][1] t
										deleteKeys bipObj[3][1].keys #selection
									)
									if horizontalIndex == 0 then
									(
										selectKeys bipObj[3][2] t
										deleteKeys bipObj[3][2].keys #selection
									)
									if turningIndex == 0 then
									(
										selectKeys bipObj[3][3] t
										deleteKeys bipObj[3][3].keys #selection
									)								
								)
							)
						)
						g_LBTools_BH.f_loadSpaces a_storedSpaces
					)
					delete pt1
					delete pt2
				)
				else
				(
					messageBox "Error: Pick object aborted." title:"Biped - Align To Offset"
				)
				clearSelection()
				setCommandPanelTaskMode #motion
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on btn_alignToMatch pressed do
		(
			try
			(
				setCommandPanelTaskMode #create
				local bipObj = g_LBTools_BH.bRoot
				local pickedObj = pickObject message:"Pick a target object to align to..." rubberband:bipObj.center
				if pickedObj != undefined AND pickedObj != #escape then
				(
					select bipObj
					with animate on
					(
						local a_storedSpaces = g_LBTools_BH.f_storeSpaces()
						for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
						(
							sliderTime = t
							local verticalIndex = getKeyIndex bipObj[3][1].controller slidertime
							local horizontalIndex = getKeyIndex bipObj[3][2].controller slidertime
							local turningIndex = getKeyIndex bipObj[3][3].controller slidertime
							if verticalIndex > 0 OR horizontalIndex > 0 OR turningIndex > 0 then
							(
								at time t
								(
									bipObj.transform = pickedObj.transform
									if verticalIndex == 0 then
									(
										selectKeys bipObj[3][1] t
										deleteKeys bipObj[3][1].keys #selection
									)
									if horizontalIndex == 0 then
									(
										selectKeys bipObj[3][2] t
										deleteKeys bipObj[3][2].keys #selection
									)
									if turningIndex == 0 then
									(
										selectKeys bipObj[3][3] t
										deleteKeys bipObj[3][3].keys #selection
									)								
								)
							)
						)
						g_LBTools_BH.f_loadSpaces a_storedSpaces
					)
				)
				else
				(
					messageBox "Error: Pick object aborted." title:"Biped - Align To Match"
				)
				clearSelection()
				setCommandPanelTaskMode #motion
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on rbtns_head changed state do
		(
			if state == 1 do
			(
				bool_head = true
				bool_root = false; bool_allLimbs = false; bool_both = false; rbtns_rootLimbsBoth.state = 0; rbtns_selLimbs.state = 0; bool_selLimbs = false; bool_nextSlideKey = false; rbtns_nextSlideKey.state = 0; rbtns_lastSlideKey.state = 0
				chkbtn_leftHand.enabled = false; chkbtn_rightHand.enabled = false; chkbtn_leftFoot.enabled = false; chkbtn_rightFoot.enabled = false
				chkbtn_leftHand.state = false; chkbtn_rightHand.state = false; chkbtn_leftFoot.state = false; chkbtn_rightFoot.state = false
				a_limbSel = #()
			)
		)
		
		on rbtns_rootLimbsBoth changed state do
		(
			bool_head = false; rbtns_head.state = 0
			case rbtns_rootLimbsBoth.state of
			(
				1:(bool_root = true; bool_allLimbs = false; bool_both = false; bool_selLimbs = false; rbtns_selLimbs.state = 0; bool_nextSlideKey = false; rbtns_nextSlideKey.state = 0; rbtns_lastSlideKey.state = 0
					chkbtn_leftHand.enabled = false; chkbtn_rightHand.enabled = false; chkbtn_leftFoot.enabled = false; chkbtn_rightFoot.enabled = false
					chkbtn_leftHand.state = false; chkbtn_rightHand.state = false; chkbtn_leftFoot.state = false; chkbtn_rightFoot.state = false
					a_limbSel = #())
				2:(bool_root = false; bool_allLimbs = true; bool_both = false; bool_selLimbs = false; rbtns_selLimbs.state = 0; bool_nextSlideKey = false; rbtns_nextSlideKey.state = 0; rbtns_lastSlideKey.state = 0
					chkbtn_leftHand.enabled = false; chkbtn_rightHand.enabled = false; chkbtn_leftFoot.enabled = false; chkbtn_rightFoot.enabled = false
					chkbtn_leftHand.state = false; chkbtn_rightHand.state = false; chkbtn_leftFoot.state = false; chkbtn_rightFoot.state = false
					a_limbSel = #())
				3:(bool_root = false; bool_allLimbs = false; bool_both = true; bool_selLimbs = false; rbtns_selLimbs.state = 0; bool_nextSlideKey = false; rbtns_nextSlideKey.state = 0; rbtns_lastSlideKey.state = 0
					chkbtn_leftHand.enabled = false; chkbtn_rightHand.enabled = false; chkbtn_leftFoot.enabled = false; chkbtn_rightFoot.enabled = false
					chkbtn_leftHand.state = false; chkbtn_rightHand.state = false; chkbtn_leftFoot.state = false; chkbtn_rightFoot.state = false
					a_limbSel = #())
			)
		)
		
		on rbtns_selLimbs changed state do
		(
			if state == 1 do
			(
				bool_head = false; rbtns_head.state = 0
				bool_root = false; bool_allLimbs = false; bool_both = false; rbtns_rootLimbsBoth.state = 0; bool_selLimbs = true; bool_nextSlideKey = false; rbtns_nextSlideKey.state = 0; rbtns_lastSlideKey.state = 0
				chkbtn_leftHand.enabled = true; chkbtn_rightHand.enabled = true; chkbtn_leftFoot.enabled = true; chkbtn_rightFoot.enabled = true
				chkbtn_leftHand.state = false; chkbtn_rightHand.state = false; chkbtn_leftFoot.state = false; chkbtn_rightFoot.state = false
				a_limbSel = #()
			)
		)

		on rbtns_nextSlideKey changed state do
		(
			if state == 1 do
			(
				bool_head = false; rbtns_head.state = 0
				bool_root = false; bool_allLimbs = false; bool_both = false; rbtns_rootLimbsBoth.state = 0; bool_selLimbs = false; rbtns_selLimbs.state = 0; rbtns_lastSlideKey.state = 0; bool_nextSlideKey = true; bool_lastSlideKey = false
				chkbtn_leftHand.enabled = false; chkbtn_rightHand.enabled = false; chkbtn_leftFoot.enabled = false; chkbtn_rightFoot.enabled = false
				chkbtn_leftHand.state = false; chkbtn_rightHand.state = false; chkbtn_leftFoot.state = false; chkbtn_rightFoot.state = false
				a_limbSel = #()
			)
		)

		on rbtns_lastSlideKey changed state do
		(
			if state == 1 do
			(
				bool_head = false; rbtns_head.state = 0
				bool_root = false; bool_allLimbs = false; bool_both = false; rbtns_rootLimbsBoth.state = 0; bool_selLimbs = false; rbtns_selLimbs.state = 0; rbtns_nextSlideKey.state = 0; bool_nextSlideKey = false; bool_lastSlideKey = true;
				chkbtn_leftHand.enabled = false; chkbtn_rightHand.enabled = false; chkbtn_leftFoot.enabled = false; chkbtn_rightFoot.enabled = false
				chkbtn_leftHand.state = false; chkbtn_rightHand.state = false; chkbtn_leftFoot.state = false; chkbtn_rightFoot.state = false
				a_limbSel = #()
			)
		)
		
		on chkbtn_leftHand changed state do
		(
			if state == on do
			(
				append a_limbSel g_LBTools_BH.bHandL
			)
			if state == off do
			(
				local item = findItem a_limbSel g_LBTools_BH.bHandL
				if item != 0 do
				(
					deleteItem a_limbSel item
				)
			)
		)

		on chkbtn_rightHand changed state do
		(
			if state == on do
			(
				append a_limbSel g_LBTools_BH.bHandR
			)
			if state == off do
			(
				local item = findItem a_limbSel g_LBTools_BH.bHandR
				if item != 0 do
				(
					deleteItem a_limbSel item
				)
			)
		)

		on chkbtn_leftFoot changed state do
		(
			if state == on do
			(
				append a_limbSel g_LBTools_BH.bFootL
			)
			if state == off do
			(
				local item = findItem a_limbSel g_LBTools_BH.bFootL
				if item != 0 do
				(
					deleteItem a_limbSel item
				)
			)
		)

		on chkbtn_rightFoot changed state do
		(
			if state == on do
			(
				append a_limbSel g_LBTools_BH.bFootR
			)
			if state == off do
			(
				local item = findItem a_limbSel g_LBTools_BH.bFootR
				if item != 0 do
				(
					deleteItem a_limbSel item
				)
			)
		)

		on btn_bakeAnimationKeys pressed do
		(
			try
			(
				undo "BakeAnimationKeys" on
				(
					if bool_head then
					(
						g_LBTools_BH.f_bipedBakeHead()
					)
					
					if bool_root then
					(
						if isValidNode g_LBTools_BH.bRoot.parent then
						(
							g_LBTools_BH.f_bipedBakeRoot()
						)
					)
					
					if bool_allLimbs then
					(
						g_LBTools_BH.f_bipedBakeSlidingKeys #(g_LBTools_BH.bHandL, g_LBTools_BH.bHandR, g_LBTools_BH.bFootL, g_LBTools_BH.bFootR)
					)
					
					if bool_both then
					(
						if isValidNode g_LBTools_BH.bRoot.parent then
						(
							g_LBTools_BH.a_storedSpaces = g_LBTools_BH.f_storeSpaces()
							g_LBTools_BH.f_bipedBakeRoot()
							g_LBTools_BH.f_loadSpaces g_LBTools_BH.a_storedSpaces
						)
						g_LBTools_BH.f_bipedBakeSlidingKeys #(g_LBTools_BH.bHandL, g_LBTools_BH.bHandR, g_LBTools_BH.bFootL, g_LBTools_BH.bFootR)
					)
					
					if bool_selLimbs then
					(
						if a_limbSel.count != 0 then
						(
							g_LBTools_BH.f_bipedBakeSlidingKeys a_limbSel
						)
						else
						(
							messageBox "Error: Click one or more limb buttons to process."
						)
					)
					
					if bool_nextSlideKey then
					(
						if selection.count == 1 then
						(
							local obj = selection[1]
							if isValidNode obj then
							(
								local index = findItem g_LBTools_BH.a_bipedLimbs obj
								if index != 0 then
								(
									g_LBTools_BH.f_bipedBakeSlidingKeyNext()
								)
								else
								(
									messageBox "Error: Select a single Biped limb with at least two sliding keys after the slider time." title:"Biped Bake"
								)
							)
							else
							(
								messageBox "Error: Object selected is not valid." title:"Biped Bake"
							)
						)
						else
						(
							messageBox "Error: Select a single Biped limb with at least two sliding keys after the slider time." title:"Biped Bake"
						)
					)
					
					if bool_lastSlideKey then
					(
						if selection.count == 1 then
						(
							local obj = selection[1]
							if isValidNode obj then
							(
								local index = findItem g_LBTools_BH.a_bipedLimbs obj
								if index != 0 then
								(
									g_LBTools_BH.f_bipedBakeSlidingKeyLast()
								)
								else
								(
									messageBox "Error: Select a single Biped limb with at least two sliding keys after the slider time." title:"Biped Bake"
								)
							)
							else
							(
								messageBox "Error: Object selected is not valid." title:"Biped Bake"
							)
						)
						else
						(
							messageBox "Error: Select a single Biped limb with at least two sliding keys after the slider time." title:"Biped Bake"
						)
					)
				)
			)catch(g_LBTools_BH.f_initBipedParts())
		)
		
		on pkbtn_propTarget picked obj do
		(
			pkbtn_propTarget.object = propTargetObj = obj
		)
		
		on btn_alignPropToTarget pressed do
		(
			if isValidNode propTargetObj then
			(
				if selection.count > 0 then
				(
					undo "AlignPropToTarget" on
					(
						for prop in $selection where (f_isProp prop) do
						(
							with animate on
							(
								if chkbx_propBake.checked then
								(
									for t = g_LBTools_BH.TRstart to g_LBTools_BH.TRend by g_LBTools_BH.TRstep while not keyboard.escPressed do
									(
										sliderTime = t
										prop.transform = propTargetObj.transform
									)
								)
								else
								(
									prop.transform = propTargetObj.transform
								)
							)
						)
					)
				)
				else
				(
					messagebox "Error: Select one or more props to align to the picked target." title:"Biped Bake"
				)
			)
			else
			(
				messagebox "Error: Prop target object is not valid, please pick another one." title:"Biped Bake"
				pkbtn_propTarget.object = undefined
			)
			sliderTime = animationRange.start
		)
	) --// end rollout

	rollout About_Rollout "About"
	(
		local imageGuide = (getDir #usericons) + "\LBTools_Guide_Image.bmp"
		
		label lbl_author "Author - Lorne Brooks"
		hyperLink myEmail "lbtools2k@gmail.com" across:2 align:#center offset:[30,10] color:(color 0 0 255) hovercolor:(color 0 0 190) visitedcolor:(color 190 0 0) align:#center address:"mailto:lbtools2k@gmail.com"
		button btn_guide "" align:#right offset:[-10,0] images:#(imageGuide, undefined, 1, 1, 1, 1, 1, false) tooltip:"Opens an online guide explaining the tool's features."
		label lbl_version "Version 3.2 - 13 Oct 2023"
		
		on btn_guide pressed do
		(
			shellLaunch "https://sites.google.com/view/lbtools-guide/home/biped-helper" ""
		)
	) --// end rollout
	
	fn f_openBipedUI =
	(
		g_LBTools_BH.BH_pos = mouse.screenpos + [-75,-15]
		if NOT (doesFileExist g_LBTools_BH.bhINIfile) then
		(
			setIniSetting g_LBTools_BH.bhINIfile "Floater" "Position" (g_LBTools_BH.BH_pos as string)
			setIniSetting g_LBTools_BH.bhINIfile "BipedPanel" "CurrentBip" "1"
			g_LBTools_BH.f_openBipedUI()
		)
		else
		(
			if NOT (hasIniSetting g_LBTools_BH.bhINIfile "Floater" "Position") do
			(
				setIniSetting g_LBTools_BH.bhINIfile "Floater" "Position" (g_LBTools_BH.BH_pos as string)
			)
			g_LBTools_BH.a_bips = #()
			g_LBTools_BH.f_collectBipRoots()
			if g_LBTools_BH.a_bips.count > 0 then
			(
				--// Floater definition
				try(closeRolloutFloater g_LBTools_BH.bipedUI_Floater)catch()
				g_LBTools_BH.bipedUI_Floater = newRolloutFloater "Biped Helper" 205 665
				g_LBTools_BH.BH_pos = execute(getIniSetting g_LBTools_BH.bhINIfile "Floater" "Position")
				g_LBTools_BH.bipedUI_Floater.pos = g_LBTools_BH.BH_pos
				addRollout g_LBTools_BH.SelectDisplay_Rollout g_LBTools_BH.bipedUI_Floater rolledup:false
				addRollout g_LBTools_BH.Keys_Rollout g_LBTools_BH.bipedUI_Floater rolledup:true
				addRollout g_LBTools_BH.LimbSpaces_Rollout g_LBTools_BH.bipedUI_Floater rolledup:true
				addRollout g_LBTools_BH.BipFiles_Rollout g_LBTools_BH.bipedUI_Floater rolledup:true
				addRollout g_LBTools_BH.Layers_Rollout g_LBTools_BH.bipedUI_Floater rolledup:true
				addRollout g_LBTools_BH.PosePosture_Rollout g_LBTools_BH.bipedUI_Floater rolledup:true
				addRollout g_LBTools_BH.Reference_Rollout g_LBTools_BH.bipedUI_Floater rolledup:true
				addRollout g_LBTools_BH.AlignBake_Rollout g_LBTools_BH.bipedUI_Floater rolledup:true
				addRollout g_LBTools_BH.About_Rollout g_LBTools_BH.bipedUI_Floater rolledup:true
			)
			else
			(
				messageBox "Error: No Biped's found in the scene. Add a Biped and try again." title:"Biped Helper"
				setIniSetting g_LBTools_BH.bhINIfile "BipedPanel" "CurrentBip" "1"
				try(closeRolloutFloater g_LBTools_BH.bipedUI_Floater)catch()
			)
		)
	)
	
	g_LBTools_BH.f_openBipedUI()
) --// end function

) --// end struct

g_LBTools_BH = s_LBTools_bipedHelper()


/*********************************************************************************************************************************************************************/
--// MACRO CALLS FOR UI
/*********************************************************************************************************************************************************************/

macroScript bipedHelper
	category:"LB Tools"
	buttontext:"Biped Helper 3.2"
	icon:#("LBTools_BipedHelper",1)
	tooltip:"Biped Helper 3.2"
(
	g_LBTools_BH.f_bipedHelperUI()
)

/*********************************************************************************************************************************************************************/